#include "DipDimClient.h"
#include "DipDimFactory.h"

#include "DipLogger.h"


DipDimClient::DipDimClient(const std::string & theTopicName, DipSubscriptionListener & listener, DipSubscription & subscription):
dipPublisherState(DipDimClient::Uninit), dipTypeService(NULL), dipDataService(NULL), dimListener(listener), dimSubscription(subscription), pubName(theTopicName) 
{
	// initialize the DIP quality status to GOOD... this is only for logging purposes
	lastKnownQuality = DIP_QUALITY_GOOD;

	dipTypeService = new DipTypeClient(pubName, *this, *this, *this);
	dipDataService = new DipDataClient(pubName, *this);
	dipPuStatesStrings[0] = "Un-initialized publication";
	dipPuStatesStrings[1] = "Non existing publication";
	dipPuStatesStrings[2] = "Publication is connected";
	dipPuStatesStrings[3] = "Publication structural definition";
	dipPuStatesStrings[4] = "Publication is disconnected";
	dipPuStatesStrings[5] = "Pulication is available again";
	dipPuStatesStrings[6] = "Publication structural re-definition";
}

DipDimClient::~DipDimClient(){
	if (dipTypeService){
		delete dipTypeService;
		LOG4CPLUS_DEBUG(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT("Deleted dipTypeService."));
	}
	if (dipDataService){
		delete dipDataService;
		LOG4CPLUS_DEBUG(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT("Deleted dipDataService."));
	}
}		


void DipDimClient::signalConnectionToListener(){
	try {
		dimListener.connected(&dimSubscription);
		LOG4CPLUS_INFO(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT("Subscription connected."));
	} catch(...) {
		std::string message("Uncaught exception from DipSubscriptionListener.connected()");
		LOG4CPLUS_ERROR(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT(message));
		// Add the publisher name to the exception message
		message += "  for publisher '"+ pubName+"'";
		DipException error(message);
		signalAsynchErrorToListener(error);
	}
}


void DipDimClient::signalDisconnectToListener(){
	try {
		dimListener.disconnected(&dimSubscription, "Down");
		LOG4CPLUS_INFO(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT("Subscription disconnected."));
	} catch(...) {
		std::string message("Uncaught exception from DipSubscriptionListener.disconnected()");
		LOG4CPLUS_ERROR(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT(message));
		message += "  for publisher '"+ pubName+"'";
		DipException error(message);
		signalAsynchErrorToListener(error);
	}
}


void DipDimClient::signalDipDataRxToListener(const DipData &data){
	try {
		LOG4CPLUS_DEBUG(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT("Data received and forwarded to message listener."));
		if(lastKnownQuality != data.extractDataQuality()){
			lastKnownQuality = data.extractDataQuality();
			LOG4CPLUS_INFO(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT("Subscription quality changed to '"<< lastKnownQuality<<"' (reason : '"<<data.extractReasonForQuality()<<"')"));
		}
		dimListener.handleMessage(&dimSubscription, (DipData &)data);
	} catch(...) {
		std::string message("Uncaught exception from DipSubscriptionListener.handleMessage()");
		LOG4CPLUS_ERROR(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT(message));
		message += "  for publisher '"+ pubName+"'";
		DipException error(message);
		signalAsynchErrorToListener(error);
	}
}

void DipDimClient::signalAsynchErrorToListener(DipException & de){
	try {
		dimListener.handleException(&dimSubscription, de);
		LOG4CPLUS_INFO(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT("DIP Exception raised : " << de.what()));
	} catch(...) {
		LOG4CPLUS_ERROR(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT("PANIC!: Uncaught exception from DipSubscriptionListener.handleException()"));
	}
}

void DipDimClient::dipDataReceived(const DipData & data){

	if ((dipPublisherState == PubDoesNotExist) || 
		(dipPublisherState == Uninit)){
		signalConnectionToListener();
		// this data is originating from the type service
		dipPublisherState = PubConnected;
	}
	else if (dipPublisherState == PubDisconnected){
		signalConnectionToListener();
		// this data is originating from the type service
		dipPublisherState = PubRestarting;
	} else if ((dipPublisherState == PubConnected) || 
			   (dipPublisherState == PubRestarting)){
		// this data is originating from the type service
		// DO NOTHING HERE - type still not defined
	} else if (dipPublisherState == PubStructDef){
		// This data is from the data service
		// DO NOTHING HERE
	} else {
		LOG4CPLUS_DEBUG(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),"Received data whilst in unexpected state "+std::string(dipPuStatesStrings[(int)dipPublisherState]));
		std::string msg("DIP data structure may have changed for subscription "+pubName);
		LOG4CPLUS_WARN(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT(msg));
		DipException de(msg);
		signalAsynchErrorToListener(de);
		return; //Do not update the listeners then.
	}

	signalDipDataRxToListener(data);
}

void DipDimClient::gotDipDataFormat(const std::string & format){
	// TODO impl as state machine
	if ((dipPublisherState == PubConnected) ||
		(dipPublisherState == PubDoesNotExist) || 
		(dipPublisherState == Uninit)){
		if ((dipPublisherState == PubDoesNotExist) || 
				(dipPublisherState == Uninit)){
				signalConnectionToListener();
		}
		try{
			cout << "Set data format " << pubName << " format "<< format << endl;
			dipDataService->setDataFormat(format);
			dipPublisherState = PubStructDef;
			LOG4CPLUS_DEBUG(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName), LOG4CPLUS_TEXT("Subscription type structure received format: "+format));
		} catch(DipException &de){
			signalAsynchErrorToListener(de);
		}
	} else if (dipPublisherState == PubRestarting){
		if (dipDataService->getDataFormat() == format){
//			cout << "Validate format " << pubName << " format "<< format << endl;
			dipDataService->validateDataFormat();
			dipPublisherState = PubStructDef;
			LOG4CPLUS_DEBUG(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName), LOG4CPLUS_TEXT("Subscription type structure received format: "+format));
		}else {
//			dipPublisherState = PubStructReDef;
//			std::string msg("DIP data structure has changed");
//			LOG4CPLUS_ERROR(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT(msg));
			// Append the subscription name for the exception
//			msg+=" for subscription '"+pubName+"'";
//			DipException de(msg);
//			signalAsynchErrorToListener(de);
// CG
			cout << "Structure changed " << pubName << " format "<< format << endl;
			try{
				dipDataService->setDataFormat(format);
				dipPublisherState = PubStructDef;
				LOG4CPLUS_DEBUG(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName), LOG4CPLUS_TEXT("Subscription type structure received format: "+format));
			} catch(DipException &de){
				signalAsynchErrorToListener(de);
			}

		}	
	} else if (dipPublisherState == PubDisconnected){
		/* OK in theory we should NOT have this transmission.
		We should go PubDisconnected -> PubRestarting -> PubStructDef
		(remember it is the publisher that has gone down and coming back
		up again). However, if the publisher is fast then it will have sent 
		data before the subscriber can make its connection - hence the client
		will not see the state where the publication data structure is undefined.
		*/
		signalConnectionToListener();
		if (dipDataService->getDataFormat() == format){
//			cout << "Validate format1 " << pubName << " format "<< format << endl;
			dipDataService->validateDataFormat();
			dipPublisherState = PubStructDef;
		}else {
//			dipPublisherState = PubStructReDef;
//			std::string msg("DIP data structure has changed for subscription "+pubName);
//			LOG4CPLUS_ERROR(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),LOG4CPLUS_TEXT(msg));
//			DipException de(msg);
//			signalAsynchErrorToListener(de);
// CG
			cout << "Structure changed1 " << pubName << " format "<< format << endl;
			try{
				dipDataService->setDataFormat(format);
				dipPublisherState = PubStructDef;
				LOG4CPLUS_DEBUG(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName), LOG4CPLUS_TEXT("Subscription type structure received format: "+format));
			} catch(DipException &de){
				signalAsynchErrorToListener(de);
			}
		}
	} else {
		// This could happen if the Publisher has redeclared its publications a clients is subscribed to,
		// before it actually sent the first piece of data.
		LOG4CPLUS_WARN(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),"Received type format string "+format+" whilst in unexpected state "+std::string(dipPuStatesStrings[(int)dipPublisherState]));
	}
}




void DipDimClient::disconnected(const DimStampedInfo & source){
	// TODO impl as state machine
	bool signalEvent = false;
	if ((dipPublisherState == PubStructDef) ||
		(dipPublisherState == PubRestarting) || 
		(dipPublisherState == PubStructReDef)){
		dipPublisherState = PubDisconnected;
		dipDataService->invalidateDataFormat();
//		cout << "Invalidate format " << pubName << endl;
		signalEvent = true;
	} else if (dipPublisherState == PubConnected){
		dipPublisherState = PubDoesNotExist;
		signalEvent = true;
	} else if (dipPublisherState == Uninit){
		dipPublisherState = PubDoesNotExist;
		// DO NOT send disconnect signal since the subscriber assumes disconnected untill a connection event occurs.
		// We have uninit state so that we can pick up a disconnect event when in PubDoesNotExist state (this should not happen)
	}else {
		LOG4CPLUS_WARN(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,pubName),"Subscription got disconnected whilst in unexpected state "+std::string(dipPuStatesStrings[(int)dipPublisherState]));
	}

	if (signalEvent){
		signalDisconnectToListener();
	}
}
