// DipDimFactory.cpp: implementation of the DipDimFactory class.
//
//////////////////////////////////////////////////////////////////////
#ifdef WIN32
#pragma warning(disable:4786)
#endif
#include "DipDimFactory.h"
#include "DipDimPublication.h"
#include "DipDimBrowser.h"
#include "DipSyntaxEnforcer.h"
#include "DIPVersion.h"
//#include "dis.hxx"

#include "DipLogger.h"

using namespace std;


DipDimFactory::DipDimFactory(const char *username, const char* logPropertiesFile, int loggingConfigWatchInterval):
dimServerStarted(false), //DIM server not yet started indeed.
lock("DipFactoryLock")
	{
	log4cplusConfigureAndWatchThread = DipLogger::initLogging(logPropertiesFile, loggingConfigWatchInterval);

	if(username == NULL)
		dimServerName = "";
	else
		{
		dimServerName = username;
		}
	//Set DIM timeout after fill up of sending socket buffer and before closing the connection:20s.
	dim_set_write_timeout(20);
	//Disable DIM padding.
	dic_disable_padding();
	dis_disable_padding();
	//Check whether env. variables have overriden the hardcoded default.
	char * dnsNames = getenv(defaultDnsEnvVarName);
	if (dnsNames == NULL){
		DimClient::setDnsNode(defaultDnsList, defaultDnsPort);
		DimServer::setDnsNode(defaultDnsList, defaultDnsPort);
		}
	else{	
		DimClient::setDnsNode(dnsNames, defaultDnsPort);
		DimServer::setDnsNode(dnsNames, defaultDnsPort);
		}
	//Running the DIM main loop. All next publications and subscriptions shall be done on the provided DNS.
	DimServer::autoStartOff();


	}

void DipDimFactory::setName(const char *username)
	{
	if(dimServerName.size() != 0){
		dimServerName = "";
		}

	if(username != 0){
		dimServerName = username;
		}
	}

DipDimFactory::~DipDimFactory()
	{
	DimServer::stop();
		if(log4cplusConfigureAndWatchThread != NULL){
			delete log4cplusConfigureAndWatchThread;
		}
	}

/*
Used to override the default. 
*/
void DipDimFactory::setDNSNode(const char* nodename) {

	if (NULL == nodename || string("").compare(nodename)== 0) {
		LOG4CPLUS_WARN(log4cplus::Logger::getInstance("dip.system.factory"),LOG4CPLUS_TEXT("The provided specific DNS is null, using the default ones..."));
		//DWARNING(("The provided specific DNS is null, using the default ones...\n"));
	} else {
        if (this->dimServerStarted == false) {
            LOG4CPLUS_INFO(log4cplus::Logger::getInstance("dip.system.factory"),LOG4CPLUS_TEXT("Using Dip NS nodes '"<< nodename <<"'"));

            DimClient::setDnsNode(nodename, defaultDnsPort);
            DimServer::setDnsNode(nodename, defaultDnsPort);
        } else {
                LOG4CPLUS_WARN(log4cplus::Logger::getInstance("dip.system.factory"),LOG4CPLUS_TEXT("Some publications or subscriptions have been created, the DNS cannot be changed."));
        }
	}
}

/*
Used to override the standard timeout.
c.f. DIP-67
*/
void DipDimFactory::setTimeout(const int timeout) {
	int timeoutDefault;
	if(timeout <= 2 && timeout != 0){
         DWARNING(("The provided timeout is less than 2 seconds, using 2 seconds as minimum...\n"));
		 timeoutDefault = 2;
		}else{
			 timeoutDefault = timeout;
			if (this->dimServerStarted == false) {
				  dim_set_write_timeout(timeoutDefault);
				  DWARNING(("New DIM write timeout is %d s ",dim_get_write_timeout()));
				} else {
					DWARNING(("Some publications or subscriptions have been created, the DNS cannot be changed.\n"));
				}
		}
}

int DipDimFactory::getTimeout() {
  return dim_get_write_timeout();
}

DipSubscription* DipDimFactory::createDipSubscription(const char *topicname, DipSubscriptionListener *listener) 
	{
	//Lazy initialization: Start DIM server as late as possible.
	if (false == dimServerStarted) {
		std::string dipVersionServiceName = dimServerName+"/DIP_VERSION";
		dipVersionService = new DimService(const_cast<char *>(dipVersionServiceName.c_str()), const_cast<char *>(DipVersion::getDipVersion().c_str()));
		DimServer::start(dimServerName.c_str());
		dimServerStarted = true;
		LOG4CPLUS_INFO(log4cplus::Logger::getInstance("dip.system.factory"),"Started DIM Server with name "<< dimServerName);
	}

	if (topicname == NULL){
		throw DipException("Must supply publication name when creating subscription");
		}
	if (listener == NULL){
		throw DipException("Must supply listener when creating subscription");
		}

	//Check syntax of publication's name, only if the logger is enabled.
    DipSyntaxEnforcer::checkPublicationName(topicname, DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX, topicname, ".syntaxchecker"));

	DipSubscription *dipSubscription = NULL;
	dipSubscription = new DipDimSubscription(std::string(topicname), *listener);
//	DINFO(("Created subscription to %s\n", topicname));
	LOG4CPLUS_INFO(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX, topicname),LOG4CPLUS_TEXT("Created subscription '" << topicname<< "'"));
//	m_diagnosticPTree.put(DipLogger::getDotSeparatedTopicName(LOG_DIP_SUB_PREFIX, topicname,".active"),"1");
	return dipSubscription;
	}

DipPublication* DipDimFactory::createDipPublication(const char *topicname, DipPublicationErrorHandler * handler) 
	{
	//Lazy initialization: Start DIM server as late as possible.
	if (false == dimServerStarted) {
		std::string dipVersionServiceName = dimServerName+"/DIP_VERSION";
		dipVersionService = new DimService(const_cast<char *>(dipVersionServiceName.c_str()), const_cast<char *>(DipVersion::getDipVersion().c_str()));
		DimServer::start(dimServerName.c_str());
		LOG4CPLUS_INFO(log4cplus::Logger::getInstance("dip.system.factory"),"Started DIM Server with name "<< dimServerName);
		dimServerStarted = true;
		}

	if (topicname == NULL){
		throw DipException("Must supply publication name when creating publication");
		}
	if (handler == NULL){
		throw DipException("Must supply error handler when creating publication");
		}

	if(dimServerName.size() == 0){
		throw DipException(std::string("Server name is undefined"));
		}

	DipPublication *dipPublication = NULL;

	//Check syntax of publication's name, if logger is enabled
	log4cplus::Logger logger( DipLogger::getLoggerFromTopic(LOG_DIP_PUB_PREFIX, topicname, ".syntaxchecker") );
	if(logger.isEnabledFor(log4cplus::WARN_LOG_LEVEL)){
	  DipSyntaxEnforcer::checkPublicationName(topicname, logger);
	}

		{
			// This is a critical section - should not overlap with other threads or calls to destroyDipPublication()
			SmutexGuard guard(lock);
			std::map<const char *, const DipPublication *, CStrLess>::iterator it =  pubList.find(topicname);
			if (it != pubList.end()){
				throw DipException(std::string("Publication name ") + topicname + " already defined");
			}

			std::string name(topicname);
			dipPublication = new DipDimPublication(name, *handler);
			pubList[dipPublication->getTopicName()] = dipPublication;
//			DINFO(("Created publication %s\n", topicname));
			LOG4CPLUS_INFO(DipLogger::getLoggerFromTopic(LOG_DIP_PUB_PREFIX, topicname),LOG4CPLUS_TEXT("Created publication '" << topicname<< "'"));
//			m_diagnosticPTree.put(DipLogger::getDotSeparatedTopicName(LOG_DIP_PUB_PREFIX, topicname,".active"),1);
		}
		return dipPublication;
	}

DipData* DipDimFactory::createDipData() 
	{
	DipData *dipData = new DipDimData();
	return dipData;
	}

void DipDimFactory::destroyDipSubscription(DipSubscription *subscription) 
	{
	DTRACE(("In destroyDipSubscription\n"));

	if (subscription == NULL){
		throw DipException("DipDimFactory::destroyDipSubscription - Usr passed NULL as subscription ptr");
		}
	LOG4CPLUS_INFO(DipLogger::getLoggerFromTopic(LOG_DIP_SUB_PREFIX,  subscription->getTopicName()),LOG4CPLUS_TEXT("Destroying subscription '" << subscription->getTopicName()<< "'"));
//	DINFO(("Destroying subscription %s\n", subscription->getTopicName()));

	delete (DipDimSubscription *)subscription;
//	m_diagnosticPTree.put(DipLogger::getDotSeparatedTopicName(LOG_DIP_SUB_PREFIX, subscription->getTopicName(),".active"),0);

	DTRACE(("Out destroyDipSubscription\n"));
	}

void DipDimFactory::destroyDipPublication(DipPublication *publication) 
	{
//	DTRACE(("In destroyDipPublication\n"));

	if (!publication){
		throw DipException("DipDimFactory::destroyDipPublication - Usr passed NULL as publication ptr");
		}

		{
		// This is a critical section - should not overlap with other threads or calls to createDipPublication()
		SmutexGuard guard(lock);
		std::map<const char *, const DipPublication *, CStrLess>::iterator it =  pubList.find(publication->getTopicName());
		if (it == pubList.end()){
			throw DipInternalError(std::string("Publication unknown to factory - ") + publication->getTopicName());
			}

		pubList.erase(it);
		LOG4CPLUS_INFO(DipLogger::getLoggerFromTopic(LOG_DIP_PUB_PREFIX,  publication->getTopicName()),LOG4CPLUS_TEXT("Destroying publication '" << publication->getTopicName()<< "'"));
//		DINFO(("Destroying publication %s\n", publication->getTopicName()));
		delete (DipDimPublication *)publication;
//		m_diagnosticPTree.put(DipLogger::getDotSeparatedTopicName(LOG_DIP_PUB_PREFIX, publication->getTopicName(),".active"),0);
		}
//		DTRACE(("Out destroyDipPublication\n"));
	}



DipBrowser* DipDimFactory::createDipBrowser(){
//	DTRACE(("In/out DipDimFactory::createDipBrowser\n"));
	return new DipDimBrowser();
	}

//Initializing the DipDimFactory constants.
const char* DipDimFactory::defaultDnsList = "DIPNSDEV.cern.ch";
const int DipDimFactory::defaultDnsPort = 2506;
const char* DipDimFactory::defaultDnsEnvVarName = "DIPNS";

