#ifndef dipdimdata_h
#define dipdimdata_h
//#pragma warning(disable:4786)

#include "Options.h"
#include <map>
#include "Field.h"
#include "DipData.h"
#include "DipDataType.h"

class DipDataExtractor;
class DipDataInserter;



/**
* function used for ordering entries in the maps. For some reason less<>
* was not working.
*/
struct StringLess  {
	bool operator()(const std::string * _X, const std::string * _Y) const
		{return (*_X < *_Y); 
	}
};



/**
 * Interface for DIP data. DipData objects are created by DipFactoryInterfaces and
 * returned to DipSubscriptionListeners.
 * For the time being we deal only with single value per data object. A extension by
 * which named data values can be inserted and extracted is foreseen.
 */
class DipDimData:public DipData
{
private:
	/**
	* used to indicate when the fields have changed. Next call to getTags()
	* will check this and rebuild list if true;
	*/
	bool fieldsChanged;


	/// Quality of the data - used on the client side
	DipQuality dipQuality;

	/// Only used on the client side
	std::string reasonForQuality;
	
	/// Only used on the client side  
	DipTimestamp dipTimestamp;
	   

	/**
	* Holds the field names.
	* This member exists so that we have a consistent DiData interface - 
	* All arrays passed to the user are still owned by the DipData Object
	*/
	char ** fieldNamesHolder;

	//private boolean isPrimitive;
	//private boolean isComplex;
	

	/**
	* Find the field (or NULL) with the specified tag name
	*/
	Field * findField(const char tag[]) const;
	
	/**
	* Get specified tag - throws eexcption if not valid
	*/
	Field & retrieveField(const char tag[]) const; //throws BadParameter

	
	/**
	* Get a field object - throws an exception if does not exist or of wrong type
	*/
	Field & retrieveTypedField(const char tag[], const DipDataType type) const;//throws BadParameter, TypeMismatch



	/**
	* Gets a field, will throw an exception if of wrong type
	* @return Field - not owned, may be NULL
	*/
	Field * lookForTypedField(const char tag[], const DipDataType type) const;// throws TypeMismatch;
	
	
	/**
	* Gets a field, will throw an exception if of wrong type or cardinality
	* @return Field - not owned, may be NULL
	*/
	Field *lookForTypedField(const char tag[], const DipDataType type, const unsigned int cardinality) const;// throws TypeMismatch;


	
	
protected:
	/// Map of fields - OWNED!
	std::map<const std::string *, Field *, StringLess> fields;

	
public:
	/// field name used when dip primitive is sent
	static const char *fieldNameWhenPrimitive;


	/**
	 * Data object created with a default data quality of good
	 * */
	DipDimData():dipQuality(DIP_QUALITY_GOOD), reasonForQuality(""), 
		fieldNamesHolder(NULL)
	{
		// force constuction of list if no elements have been added
		fieldsChanged = true;
	};


	
	virtual ~DipDimData();

	

	const std::map<const std::string *, Field *, StringLess> & getFields() const{
		return fields;
	}
	
	
	void setDataQuality(DipQuality dataQuality){
		dipQuality = dataQuality;
	}


	void setDipTime(const DipTimestamp & time){
		dipTimestamp = time;
	}

	void setQualityString(const std::string &reason){
		reasonForQuality = reason;
	}


	const std::string &	extractString	(const char *tag=NULL) const;
	const char *		extractCString	(const char *tag=NULL) const;
	DipBool				extractBool		(const char *tag=NULL) const;
	DipByte				extractByte		(const char *tag=NULL) const;
	DipShort			extractShort	(const char *tag=NULL) const;
	DipInt				extractInt		(const char *tag=NULL) const;
	DipFloat			extractFloat	(const char *tag=NULL) const;
	DipDouble			extractDouble	(const char *tag=NULL) const;
	DipLong				extractLong		(const char *tag=NULL) const;
	const DipBool *		extractBoolArray(int &size,	const char *tag=NULL) const;
	const DipByte *		extractByteArray(int &size, const char *tag=NULL) const;
	const DipShort * 	extractShortArray(int &size, const char *tag=NULL) const;
	const DipInt *		extractIntArray	(int &size,	const char *tag=NULL) const;
	const DipFloat * 	extractFloatArray(int &size, const char *tag=NULL) const;
	const DipDouble * 	extractDoubleArray(int &size,const char *tag=NULL) const;
	const DipLong *		extractLongArray(int &size,	const char *tag=NULL) const;
	const std::string *	extractStringArray(int &size,const char *tag=NULL) const;
	const char **		extractCStringArray(int &size,const char *tag=NULL) const;


	/**
	* Generic routine to insert field and its data
	*/
	void insert(const void *data, const char tag[], const DipDataType type, const int cardinality);
	void insert(const std::string &value,	const char *tag=NULL);
	void insert(const char value[],	const char *tag=NULL);
	void insert(DipBool value,		const char *tag=NULL);
	void insert(DipByte value,		const char *tag=NULL);
	void insert(DipShort value,		const char *tag=NULL);
	void insert(DipInt value,		const char *tag=NULL);
	void insert(DipLong value,		const char *tag=NULL);
	void insert(DipFloat value,		const char *tag=NULL);
	void insert(DipDouble value,	const char *tag=NULL);
	void insert(const DipBool addr[],	int size, const char *tag=NULL);
	void insert(const DipByte addr[],	int size, const char *tag=NULL);
	void insert(const DipShort addr[],	int size, const char *tag=NULL);
	void insert(const DipInt addr[],	int size, const char *tag=NULL);
	void insert(const DipLong addr[],	int size, const char *tag=NULL);
	void insert(const DipFloat addr[],	int size, const char *tag=NULL);
	void insert(const DipDouble addr[],	int size, const char *tag=NULL);
	void insert(const std::string value[],int size,const char *tag=NULL);
	void insert(const char *value[],	int size, const char *tag=NULL);

	DipDataType getValueType(const char *tag=NULL) const;

	/**
	* Recalculated if fieldsChanged is true;
	* 
	*/
	const char **getTags(int &nTags) const;
	bool contains(const char tag[]) const;
	unsigned int size() const;
	bool isEmpty() const;

	DipQuality	extractDataQuality() const;
	const std::string & extractReasonForQuality() const;
	const DipTimestamp & extractDipTime() const;

	/**
	* returns the size of the supplied type. If the type is an array, the size
	* of one of its elements is returned. If the type is a string 1 is returned
	*/
	static int getTypeSize(DipDataType dipType);


	/**
	* The dimension of the field - 0 if field does not exist
	*/
	unsigned getValueDimension(const char *tag) const;

	/**
	* Returns the byte size of the data held in each field
	* (all string null terminators are included).
	*/
	unsigned int dataByteSize() const;
};



#endif

