#ifndef Service_DIPDataMapper_h
#define Service_DIPDataMapper_h
#include "Options.h"
#include "Dip.h"
#include "Field.h"
#include "DipException.h"
#include <string>
#include <vector>


/**
 * Holds information on the fields offset within the DIM service
 * @see Field
 * */
class  DimServiceField{
public:
	///Obtained from the first send - OWNED
	Field &dipField;

	/// byte offset in the DIM service where the start of the data for this field can be found
	int serviceOffset; 
	
	DimServiceField(Field *field, int offset):dipField(*field){
		serviceOffset = offset;
	}

	~DimServiceField(){
		delete &dipField;
	}
};



/**
 * @author beharrel
 *
 * Used in the serialising and deserialising of DIP data to from DIM services.
 * The DIM service is acting as a large data buffer that is split into three parts.
 * <ol>
 * <li> Fixed, holds primitives and arrays of primitives</li>
 * <li> Varable, holds strings and arrays of strings</li>
 * <li> A quality reason string - this is not part of the users DIP data structure, 
 * rather an attribute associated with the quality, hence it will NOT appear in the structure description</li>
 * </ol>
 * 
 */
class DipDllExp Service_DIPDataMapper {
private:
	void splitFormatStringIntoFields(std::string &formatString, std::vector<std::string> & stringFields);

protected:

	/**
	 * String that describes the format of the DIP Data. It is published by the Type service so that the client 
	 * side knows how to deserialse the data received via the data service.
	 * structure string is of the format:-</p>
	 * [<tagName>|<byteOffset>|numBytes|<enumeratedType>|]...</p>
	 * Where the enumeratedType is a primitive of array of primitives the offset points to the offset in the fixed part of the
	 * service buffer and number of bytes refers to the size of the field in bytes.
	 * Where the enumeratedType is TYPE_STRING or TYPE_STRING_ARRAY offset holds the sequence of that string in the variable part
	 * of the service buffer(starting from 0) and numBytes holds the cardinality of the field.
	 * */
	std::string dipDataFormatDescription;

	int dipDataFormatValid;


	/**
	* Holds the DIM description of the DIP publication
	*/
	std::string dimFormatDescription;
	
	
	/**
	 * Obtained from the First DIP data object sent. This fix's the publication type
	 * */
	std::vector<DimServiceField *> dipDataServiceFields;
	


	/// Gives the index of the first byte of the start of the variable section of the DimService
	int startOfVarSection;
	


	/// Points one past the end of the Dip data structure (EOF var part of service)
	int startOfQualityReasonString;
	
	
	
	void buildDipServiceMap();// throws DipException
// CG
	void clearDipServiceMap();// throws DipException

	
	/**
	* Used to build the DIM structure string based on the values in the Fields vector.
	* If there are no strings in the Dim structure a 'C' will be added on the end of the
	* Format string to account for the reason for Quality message.
	*/
	void buildDimFormatString();

	
	
	Field &getField(const std::string &fieldName);// throws DipException


public:
	Service_DIPDataMapper():dipDataFormatDescription(""), dipDataFormatValid(0), dimFormatDescription(""), startOfVarSection(-1), startOfQualityReasonString(-1) {};

	virtual ~Service_DIPDataMapper(){
		for (unsigned int i = 0; i < dipDataServiceFields.size(); i++){
			delete dipDataServiceFields[i];
		}
	}


	const std::string & getDimDataFormat(){
		return dimFormatDescription;
	}

	const std::string& getDipDataFormat(){
		return dipDataFormatDescription;
	}

	int getFixedSize() {
		return startOfVarSection;
	}

	bool isFormatDefined() {
		return (dipDataFormatDescription.size() != 0);
	}
	bool isFormatValid() {
		return (dipDataFormatValid);
	}


	/**
	* Returns array of copies of fieldNames
	* Caller must delete the returned array
	*/
	std::string* getFieldNames(int & noFields);// throws DipException
	
	
	
	int getFieldType(const std::string &fieldName){// throws DipException
		return getField(fieldName).fieldType;
	}
	
	
	
	int getFieldCardinality(const std::string &fieldName){// throws DipException
		return getField(fieldName).cardinality;
	}
	
};

#endif

