#include "DIPBridge.h"
#include "DipDimData.h"
#include <stdio.h>
#include <string.h>


class PublicationBridge:public DipPublicationErrorHandler{
private:
	/**
	* ref to the Java VM instance
	*/
	static JavaVM *vm;

	
	/**
	* This will give us a context through which our callbacks can access the Java VM.
	* Should only thread within which all callback instances run.
	*/
	static JNIEnv *env;


	/**
	* publication we are to send to 
	*/
	 DipPublication * pub;



	/**
	* Global reference to the instance of the java implementation of the publication 
	*/
	jobject javaPublicationInstance;
	


	/**
	* Global reference to the java implementation of the error handlers handleException
	*/
	static jmethodID javaHandleExceptionMethod;


	/**
	* Make sure we have an environment attached to the current thread
	*/
	void setUpEnv(){
		if (env == NULL){
			vm->AttachCurrentThread((void**)&env, NULL);
		}

		if (javaHandleExceptionMethod == NULL){
			jclass pubImpClass =  env->GetObjectClass(javaPublicationInstance);
			javaHandleExceptionMethod = env->GetMethodID(pubImpClass, "handleException",    "(Lcern/dip/DipException;)V");
		}
	}


public:
	PublicationBridge(JNIEnv *env, jobject pub):pub(NULL){
		env->GetJavaVM(&vm);
		javaPublicationInstance = env->NewGlobalRef(pub);
	}

	/**
	 *Can't use destructor because we need to pass the env param
	 * */
	void cleanUp(JNIEnv *env){
		DTRACE(("In PublicationBridge::cleanUp\n"));
		if (pub){
			DTRACE(("publicationBridge::cleanUp - about to destroy pub\n"));
			DipJavaCPPBridge::getFactory().destroyDipPublication(pub);
			DTRACE(("PublicationBridge::cleanUp - Destroyed pub\n"));
		}
		DTRACE(("PublicationBridge:cleanUp - About to del global ref\n"));
		env->DeleteGlobalRef(javaPublicationInstance);
		
		DTRACE(("Out PublicationBridge::cleanUp\n"));
	}


	void setPublication(DipPublication * publication){
		pub = publication;
	}


	/**
	* Deligates from Java to C++
	*/
	void send(const DipData& value,	longlong time_ns){
		send(value, time_ns, DIP_QUALITY_GOOD, "");
	}

	void send(const DipData& value,	longlong time_ns, const DipQuality quality, const char * reason){
		DipTimestamp ts;
		ts.setNanos(time_ns);
		pub->send(value, ts, quality, reason);
	}

	
	/**
	*Deligates from Java to C++
	*/
	void setQualityBad(const char * reason) throw (DipException){
		if (reason){
			pub->setQualityBad(reason);
		} else {
			pub->setQualityBad();
		}
	}
	

	/**
	*Deligates from Java to C++
	*/
	void setQualityUncertain(const char *reason) throw (DipException){
		if (reason){
			pub->setQualityUncertain(reason);
		} else {
			pub->setQualityUncertain();
		}
	}


	/**
	*Deligates from C++ to Java
	*/
	void handleException(DipPublication* publication, DipException& ex){
		setUpEnv();

		static jclass dipExceptionClass = env->FindClass("cern/dip/DipException");
		static jmethodID exceptionConstructor = env->GetMethodID(dipExceptionClass,"<init>", "(Ljava/lang/String;)V");
		
		// TODO expand on types of exception passed throgh to java
		jstring jreason = env->NewStringUTF(ex.what());
		jobject javaException = env->NewObject(dipExceptionClass, exceptionConstructor, jreason);

		env->CallVoidMethod(javaPublicationInstance, javaHandleExceptionMethod, javaException);
		if (env->ExceptionOccurred()){
			DERROR(("Exception detected on return from Java implementation of DipPublicationErrorHandler::handleException() - this should have been caught in the Java implementation\n"));
			env->ExceptionClear();
		}

		env->DeleteLocalRef(javaException);
		env->DeleteLocalRef(jreason);
	}
};





JavaVM * PublicationBridge::vm = NULL;
JNIEnv * PublicationBridge::env = NULL;
jmethodID PublicationBridge::javaHandleExceptionMethod=NULL;





static CdipObjectPool<PublicationBridge> pubPool("publisherPool");
static SMutex publock("PublisherImpLock");



JNIEXPORT jint JNICALL Java_cern_dip_implementation_DipPublicationImp__1create
  (JNIEnv *env, jobject pub, jstring pubName){
	SmutexGuard guard(publock);
	DTRACE(("In DipPublicationImp__1create\n"));

	const char * pubName_cstr = pubName ? env->GetStringUTFChars(pubName, NULL) : NULL;
	PublicationBridge * pubBridge = new PublicationBridge(env, pub);
	DipPublication * publication = NULL;
	int objectIndex = -1;
	try{
		publication = DipJavaCPPBridge::getFactory().createDipPublication(pubName_cstr,pubBridge);
		pubBridge->setPublication(publication);
		objectIndex = pubPool.give(pubBridge);
		if (pubName_cstr){
			env->ReleaseStringUTFChars(pubName, pubName_cstr);
		}
	} catch (DipException &de){
		delete pubBridge;
		if (pubName_cstr){
			env->ReleaseStringUTFChars(pubName, pubName_cstr);
		}
		DipJavaCPPBridge::flagJavaDIPException(env, de.what());
	}

	
	DTRACE(("Out DipPublicationImp__1create\n"));
	return objectIndex;
}



JNIEXPORT void JNICALL Java_cern_dip_implementation_DipPublicationImp__1destroy
(JNIEnv * env, jobject, jint objIndex){
	SmutexGuard guard(publock);
	DTRACE(("In DipPublicationImp__1destroy\n"));

	PublicationBridge * bridge = pubPool.remove(objIndex);
	bridge->cleanUp(env);
	delete bridge;

	DTRACE(("Out DipPublicationImp__1destroy\n"));
}




JNIEXPORT void JNICALL Java_cern_dip_implementation_DipPublicationImp__1setQualityBad
(JNIEnv * env, jobject, jint objIndex, jstring reason){
	SmutexGuard guard(publock);
	DTRACE(("In DipPublicationImp__1setQualityBad\n"));

	const char * reason_cstr = reason ? env->GetStringUTFChars(reason, NULL) : NULL;
	try{
		pubPool.retrieve(objIndex).setQualityBad(reason_cstr);
	} catch (DipException &de){
		DipJavaCPPBridge::flagJavaDIPException(env, de.what());
	}
	
	if (reason_cstr){
		env->ReleaseStringUTFChars(reason, reason_cstr);
	}
	DTRACE(("Out DipPublicationImp__1setQualityBad\n"));
}



JNIEXPORT void JNICALL Java_cern_dip_implementation_DipPublicationImp__1setQualityUncertain
  (JNIEnv *env, jobject, jint objIndex, jstring reason){
	SmutexGuard guard(publock);
	DTRACE(("In DipPublicationImp__1setQualityUncertain\n"));

	const char * reason_cstr = reason ? env->GetStringUTFChars(reason, NULL) : NULL;
	try {
		pubPool.retrieve(objIndex).setQualityUncertain(reason_cstr);
	}catch (DipException &de){
		DipJavaCPPBridge::flagJavaDIPException(env, de.what());
	}

	if (reason_cstr){
		env->ReleaseStringUTFChars(reason, reason_cstr);
	}
	DTRACE(("Out DipPublicationImp__1setQualityUncertain\n"));
}



/**
* Write one field to the DipData object
*/
void writeToDataObject(DipDimData & dataObject, char **bufferPtr, const char * fieldName, int type, int cardinality){
	switch (type){
	case TYPE_BOOLEAN:{
						DipBool val = **((DipBool **)bufferPtr);
						dataObject.insert(val, fieldName);
						*bufferPtr += sizeof(val);
						break;
					  }
	case TYPE_BYTE:{
						DipByte val = **((DipByte **)bufferPtr);
						dataObject.insert(val, fieldName);
						*bufferPtr += sizeof(val);
						break;
					}
	case TYPE_SHORT:{
						DipShort val = **((DipShort **)bufferPtr);
						dataObject.insert(val, fieldName);
						*bufferPtr += sizeof(val);
						break;
					}
	case TYPE_INT:{
						DipInt val = **((DipInt **)bufferPtr);
						dataObject.insert(val, fieldName);
						*bufferPtr += sizeof(val);
						break;
					}
	case TYPE_LONG:{
						DipLong val = **((DipLong **)bufferPtr);
						dataObject.insert(val, fieldName);
						*bufferPtr += sizeof(val);
						break;
					}
	case TYPE_FLOAT:{
						DipFloat val = **((DipFloat **)bufferPtr);
						dataObject.insert(val, fieldName);
						*bufferPtr += sizeof(val);
						break;
					}
	case TYPE_DOUBLE:{
						DipDouble val = **((DipDouble **)bufferPtr);
						dataObject.insert(val, fieldName);
						*bufferPtr += sizeof(val);
						break;
					}
	case TYPE_STRING:
					{
						const char *val = *bufferPtr;
						int len = strlen(val);
						dataObject.insert(val, fieldName);
						*bufferPtr += len + 1;
						break;
					}
	case TYPE_BOOLEAN_ARRAY:{
						DipBool *val = *(DipBool **)bufferPtr;
						int len = sizeof(*val) * cardinality;
						dataObject.insert(val, cardinality, fieldName);
						*bufferPtr += len;
						break;
					}
	case TYPE_BYTE_ARRAY:{
						DipByte *val = *(DipByte **)bufferPtr;
						int len = sizeof(*val) * cardinality;
						dataObject.insert(val, cardinality, fieldName);
						*bufferPtr += len;
						break;
					}
	case TYPE_SHORT_ARRAY:{
						DipShort *val = *(DipShort **)bufferPtr;
						int len = sizeof(*val) * cardinality;
						dataObject.insert(val, cardinality, fieldName);
						*bufferPtr += len;
						break;
					}
	case TYPE_INT_ARRAY:{
						DipInt *val = *(DipInt **)bufferPtr;
						int len = sizeof(*val) * cardinality;
						dataObject.insert(val, cardinality, fieldName);
						*bufferPtr += len;
						break;
					}
	case TYPE_LONG_ARRAY:{
						DipLong *val = *(DipLong **)bufferPtr;
						int len = sizeof(*val) * cardinality;
						dataObject.insert(val, cardinality, fieldName);
						*bufferPtr += len;
						break;
					}
	case TYPE_FLOAT_ARRAY:{
						DipFloat *val = *(DipFloat **)bufferPtr;
						int len = sizeof(*val) * cardinality;
						dataObject.insert(val, cardinality, fieldName);
						*bufferPtr += len;
						break;
					}
	case TYPE_DOUBLE_ARRAY:{
						DipDouble *val = *(DipDouble **)bufferPtr;
						int len = sizeof(*val) * cardinality;
						dataObject.insert(val, cardinality, fieldName);
						*bufferPtr += len;
						break;
					}
	case TYPE_STRING_ARRAY:{
						char ** strArray = new char * [cardinality];
						for (int strArrayInd = 0; strArrayInd < cardinality; strArrayInd++){
							strArray[strArrayInd] = *bufferPtr;
							int len = strlen(*bufferPtr);
							*bufferPtr += len + 1;
						}
						dataObject.insert((const char **)strArray, cardinality, fieldName);
						delete [] strArray;
						break;
					}
	}
}



JNIEXPORT void JNICALL Java_cern_dip_implementation_DipPublicationImp__1send
(JNIEnv *env, jobject, jint objectIndex, jobject buffer, jstring format, jint noFields, jlong time_ns, jint dipQuality, jstring qualityReason){
	SmutexGuard guard(publock);
	DTRACE(("In DipPublicationImp__1send\n"));

	int noDescriptorParts = noFields * 3;
	const char **fieldDescriptions = new const char * [noDescriptorParts];
	const char * cFormat = env->GetStringUTFChars(format, 0);
	const char * cQualityReason = env->GetStringUTFChars(qualityReason, 0);
	// make mutable copy of format string
	int formatLen = strlen(cFormat);
	char * formatCpy = new char[formatLen];
	for (int i = 0; i < formatLen; i++){
		if (cFormat[i] == '|'){
			formatCpy[i] = 0;
		}else{
			formatCpy[i] = cFormat[i];
		}
	}
	assert(formatCpy[formatLen - 1] == 0);
	const char * part = formatCpy;
	
	// seperate each part of the format description string
	for (int fieldDescIndex = 0; fieldDescIndex < noDescriptorParts; fieldDescIndex++){
		fieldDescriptions[fieldDescIndex] = part;
		int len = strlen(part);
		part += len + 1; // point to beginning of next part
	}

	// OK now build the Data object
	DipDimData * dataObject = (DipDimData *)DipJavaCPPBridge::getFactory().createDipData();
	char * bufferPtr = (char *)env->GetDirectBufferAddress(buffer); 
	try{
		for (int fieldDescIndex = 0; fieldDescIndex < (noFields * 3); fieldDescIndex+=3){
			int type, cardinality;
			const char * fieldName = fieldDescriptions[fieldDescIndex];
			sscanf(fieldDescriptions[fieldDescIndex+1],"%d",&type);
			sscanf(fieldDescriptions[fieldDescIndex+2],"%d",&cardinality);

			writeToDataObject(*dataObject, &bufferPtr, fieldName, type, cardinality);
		}

		pubPool.retrieve(objectIndex).send(*dataObject,	time_ns, (DipQuality) dipQuality, cQualityReason);
	}catch (DipException &de){
		DipJavaCPPBridge::flagJavaDIPException(env, de.what());
	}catch (std::exception &e){
		DipJavaCPPBridge::flagJavaDIPException(env, e.what());
	}catch (...){
		DipJavaCPPBridge::flagJavaDIPException(env, "Unknown source of exception");
	}

	delete dataObject;
	env->ReleaseStringUTFChars(format, cFormat);
	env->ReleaseStringUTFChars(qualityReason, cQualityReason);
	delete [] fieldDescriptions;
	delete [] formatCpy;
	DTRACE(("Out DipPublicationImp__1send\n"));
}
