#ifdef WIN32
#pragma warning ( disable: 4786 )
#endif
#include "DIPBridge.h"
#include "DipDimData.h"
#include "DipSubscription.h"
#include <stdio.h>
#include <string.h>


/**
* Used to pass callacks though to the Java implementation.
*/
class SubscriptionBridge:public DipSubscriptionListener{
private:
	/**
	* ref to the Java VM instance
	*/
	static JavaVM *vm;

	/**
	* This will give us a context through which our callbacks can access the Java VM.
	* Should only thread within which all callback instances run.
	*/
	static JNIEnv *env;


	/**
	* subscription we are recving from - owned 
	*/
	DipSubscription * sub;

	/**
	* Global reference to the instance of the java implementation of the subscription
	*/
	jobject javaSubscriptionInstance;

	/**
	* the ID of the handleMessage method of the Java Subscription implementation
	*/
	static jmethodID handleMessageMethod;


	/**
	* the ID of the connected method of the Java Subscription implementation
	*/
	static jmethodID connectedMethod;

	
	/**
	* the ID of the disconnected method of the Java Subscription implementation
	*/
	static jmethodID disconnectedMethod;


	/**
	* the ID of the handleException method of the Java Subscription implementation
	*/
	static jmethodID handleExceptionMethod;


	/**
	* Make sure we have an environment attached to the current thread
	*/
	void setUpEnv(){
		if (env == NULL){
			vm->AttachCurrentThread((void**)&env, NULL);
			jclass subImpClass   = env->GetObjectClass(javaSubscriptionInstance);
			handleMessageMethod  = env->GetMethodID(subImpClass, "handleMessage",	"(Ljava/lang/String;Ljava/nio/ByteBuffer;ILjava/lang/String;J)V");
			connectedMethod      = env->GetMethodID(subImpClass, "connected",		"()V");
			disconnectedMethod	 = env->GetMethodID(subImpClass, "disconnected",	"(Ljava/lang/String;)V");
			handleExceptionMethod= env->GetMethodID(subImpClass, "handleException",	"(Ljava/lang/Exception;)V");
		}
	}


public:
	SubscriptionBridge(JNIEnv *env, jobject jsub):sub(NULL){
		if (vm == NULL){
			env->GetJavaVM(&vm);
		}

		javaSubscriptionInstance = env->NewGlobalRef(jsub);
	}

	/**
	* can not use destructor since we need to pass the env param
	*/
	void cleanUp(JNIEnv *env){
		if (sub){
			DipJavaCPPBridge::getFactory().destroyDipSubscription(sub);
		}
		env->DeleteGlobalRef(javaSubscriptionInstance);
	}



	void setSubscription(DipSubscription * subscription){
		sub = subscription;
	}


	

	/**
	* Build data buffer to send to the Java implementation
	*/
	void handleMessage(DipSubscription* subscription, DipData& message){
		DTRACE(("In SubscriptionBridge::handleMessage\n"));

		setUpEnv();
		DipDimData & data = (DipDimData &) message;
		const unsigned int byteSize = data.dataByteSize();
		char * rawBuffer = new char[byteSize];

		jobject byteBuffer = env->NewDirectByteBuffer((void*)rawBuffer, byteSize);
		const std::map<const std::string *, Field *, StringLess> & fieldMap = data.getFields();
		std::map<const std::string *, Field *, StringLess>::const_iterator fieldIt= fieldMap.begin();
		std::string format = "";
		char * writePoint = rawBuffer;

		// Make data buffer and format string for passing to java implementation
		// Note field stores all data (str arrays) as contigious blocks of memory
		// this make our life easier here!
		while (fieldIt != fieldMap.end()){
			Field & field = *(fieldIt->second);
			char type[5], cardinality[10];
			sprintf(type, "%d", field.fieldType);
			sprintf(cardinality, "%d", field.cardinality);
			format += field.fieldName + '|' + type + '|' + cardinality + '|';
			memcpy(writePoint, field.data, field.byteSize);
			writePoint+=field.byteSize;
			fieldIt++;
		}

		jstring jformat = env->NewStringUTF(format.c_str());
		jint jquality = message.extractDataQuality();
		jstring jqualityReason = env->NewStringUTF(message.extractReasonForQuality().c_str());
		jlong jtime_ns = message.extractDipTime().getAsNanos();
		// Make the call
		env->CallVoidMethod(javaSubscriptionInstance, handleMessageMethod, jformat, byteBuffer, jquality, jqualityReason, jtime_ns);
		if (env->ExceptionOccurred()){
			DERROR(("Exception detected on return from Java implementation of SubscriptionListener::handleMessage() - this should have been caught in the Java implementation\n"));
			env->ExceptionClear();
		}

		env->DeleteLocalRef(byteBuffer);
		env->DeleteLocalRef(jformat);
		env->DeleteLocalRef(jqualityReason);
		delete[] rawBuffer;

		DTRACE(("Out SubscriptionBridge::handleMessage\n"));
	}


	/**
	* Pass the disconnection event to the Java DipSubscription implementation 
	*/
	void disconnected(DipSubscription* subscription, char* reason){
		DTRACE(("In SubscriptionBridge::disconnected\n"));

		setUpEnv();
		jstring jreason = env->NewStringUTF(reason);
		env->CallVoidMethod(javaSubscriptionInstance, disconnectedMethod, jreason);
		if (env->ExceptionOccurred()){
			DERROR(("Exception detected on return from Java implementation of SubscriptionListener::disconnected() - this should have been caught in the Java implementation\n"));
			env->ExceptionClear();
		}
		env->DeleteLocalRef(jreason);

		DTRACE(("Out SubscriptionBridge::disconnected\n"));
	}


	/**
	* Pass the connection event to the Java DipSubscription implementation 
	*/
	void connected(DipSubscription* subscription){
		DTRACE(("In SubscriptionBridge::connected\n"));

		setUpEnv();
		env->CallVoidMethod(javaSubscriptionInstance, connectedMethod);
		if (env->ExceptionOccurred()){
			DERROR(("Exception detected on return from Java implementation of SubscriptionListener::connected() - this should have been caught in the Java implementation\n"));
			env->ExceptionClear();
		}

		DTRACE(("Out SubscriptionBridge::connected\n"));
	}



	/**
	* Delegates to the Java DipSubscription implementation 
	*/
	void handleException(DipSubscription* subscription, DipException& ex){
		DTRACE(("In SubscriptionBridge::handleException\n"));

		setUpEnv();
		static jclass dipExceptionClass = env->FindClass("cern/dip/DipException");
		static jmethodID exceptionConstructor = env->GetMethodID(dipExceptionClass,"<init>", "(Ljava/lang/String;)V");

		// TODO expand on types of exception passed through to java
		jstring jreason = env->NewStringUTF(ex.what());

		jobject javaException = env->NewObject(dipExceptionClass, exceptionConstructor, jreason);
		
		env->CallVoidMethod(javaSubscriptionInstance, handleExceptionMethod, javaException);
		if (env->ExceptionOccurred()){
			DERROR(("Exception detected on return from Java implementation of SubscriptionListener::handleException() - this should have been caught in the Java implementation\n"));
			env->ExceptionClear();
		}

		env->DeleteLocalRef(javaException);
		env->DeleteLocalRef(jreason);

		DTRACE(("Out SubscriptionBridge::handleException\n"));
	}




	/**
	* Delegates to C++ implementation
	*/
	void requestUpdate(){
		sub->requestUpdate();
	}
};




JavaVM * SubscriptionBridge::vm = NULL;
JNIEnv * SubscriptionBridge::env = NULL;
jmethodID SubscriptionBridge::handleMessageMethod;
jmethodID SubscriptionBridge::connectedMethod;
jmethodID SubscriptionBridge::disconnectedMethod;
jmethodID SubscriptionBridge::handleExceptionMethod;






/// Used to keep track of the subsscription bridges
static CdipObjectPool<SubscriptionBridge> subPool("SubscriberPool");






JNIEXPORT jint JNICALL Java_cern_dip_implementation_DipSubscriptionImp__1create
(JNIEnv *env, jobject subImp, jstring pubName){
	DTRACE(("In DipSubscriptionImp__1create\n"));

	const char * pubName_cstr = pubName ? env->GetStringUTFChars(pubName, NULL) : NULL;
	SubscriptionBridge * bridge = new SubscriptionBridge(env, subImp);
	DipSubscription * sub = DipJavaCPPBridge::getFactory().createDipSubscription(pubName_cstr, bridge);
	bridge->setSubscription(sub);

	if (pubName_cstr){
		env->ReleaseStringUTFChars(pubName, pubName_cstr);
	}

	DTRACE(("Out DipSubscriptionImp__1create\n"));
	return subPool.give(bridge);
}







JNIEXPORT void JNICALL Java_cern_dip_implementation_DipSubscriptionImp__1destroy
(JNIEnv *env, jobject, jint objectIndex){
	DTRACE(("In DipSubscriptionImp__1destroy\n"));

	SubscriptionBridge * bridge = subPool.remove(objectIndex);
	bridge->cleanUp(env);
	delete bridge;

	DTRACE(("Out DipSubscriptionImp__1destroy\n"));
}





JNIEXPORT void JNICALL Java_cern_dip_implementation_DipSubscriptionImp__1requestUpdate
(JNIEnv *, jobject, jint objectIndex){
	DTRACE(("In DipSubscriptionImp__1requestUpdate\n"));

	// TODO check for exceptions
	subPool.retrieve(objectIndex).requestUpdate();

	DTRACE(("Out DipSubscriptionImp__1requestUpdate\n"));
}
