// dipObjectPool.h: interface for the CdipObjectPool class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_DIPOBJECTPOOL_H__4EFDE2DD_315B_4D41_87FE_46B1DFBEE779__INCLUDED_)
#define AFX_DIPOBJECTPOOL_H__4EFDE2DD_315B_4D41_87FE_46B1DFBEE779__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <vector>
#include <list>
#include <assert.h>
#include "DipException.h"
#include "platformDependant.h"


template <class DIPOBJ>class CdipObjectPool  
{
private:
	/**
	* Owned by pool
	*/
	std::vector<DIPOBJ *> objects;
	std::list<int> emptyEntries;
	std::string poolName;
	SMutex lock;

public:
	CdipObjectPool(std::string name):poolName(name), lock(name+"lock"){
		DTRACE(("In & Out CdipObjectPool\n"));
	};

	virtual ~CdipObjectPool(){
		DTRACE(("In ~CdipObjectPool\n"));
		DINFO(("%d undestroyed objects remain in the pool %s\n", objects.size() - emptyEntries.size(), poolName.c_str()));
	//	for (int i = 0; i < objects.size();i++){
	//		if (objects[i] != NULL){
	//			delete objects[i];
	//		}
	//	}
		DTRACE(("Out ~CdipObjectPool\n"));
	};

	/**
	* Add object into pool
	* @param obj - ownership passed to pool
	* @return Index where object has been placed
	*/
	int give(DIPOBJ *obj){
		SmutexGuard guard(lock);
		int index = -1;
		if (emptyEntries.empty()){
			index = objects.size();
			objects.push_back(obj);
		}
		else{
			index = emptyEntries.front();
			emptyEntries.pop_front();
			assert(objects[index] == NULL);
			objects[index] = obj;
		}
		DINFO(("Inserted object to pool %s at index %d (Pool size %d)\n", poolName.c_str(), index, objects.size()));
		return index;
	}

	/**
	* Remove an object from the pool - this does NOT delete that object!
	* @return object - now owned by caller
	*/
	DIPOBJ * remove(int index){
		SmutexGuard guard(lock);
		DINFO(("Removing object from pool %s at index %d (Pool size %d)\n", poolName.c_str(), index, objects.size()));
		DIPOBJ * obj = objects[index];
		if (obj == NULL){
			throw DipInternalError("ObjectPool::remove - Object at supplied index does not exist");
		}
		objects[index] = NULL;
		emptyEntries.push_back(index);
		DINFO(("Removed object to pool %s at index %d (Pool size %d)\n", poolName.c_str(), index, objects.size()));
		return obj;
	}


	/**
	*  Retrieve an object from the pool
	*/
	DIPOBJ & retrieve(int index){
		SmutexGuard guard(lock);
		if (objects[index] == NULL){
			throw DipInternalError("ObjectPool::retrieve - Object at supplied index does not exist");
		}
		DIPOBJ * object = objects[index];
		DINFO(("Retrieved object to pool %s at index %d (Pool size %d)\n", poolName.c_str(), index, objects.size()));
		return *object;
	}
};

#endif // !defined(AFX_DIPOBJECTPOOL_H__4EFDE2DD_315B_4D41_87FE_46B1DFBEE779__INCLUDED_)
