package cern.dip;

import java.util.Map;
import java.util.Optional;
import java.io.Serializable;

/**
 * Interface for DIP data. DipData object are used by publishers to publish 
 * complex data types. These object are are {@link cern.dip.DipFactory#createDipData created} 
 * by the {@link cern.dip.DipFactory DIP} factory.
 * <p>
 * DipData objects are also delivered to
 * {@link cern.dip.DipSubscriptionListener subscription listeners}
 * as result of subscription.
 * <p>
 * DipData objects can contain one or more named entries, each holding
 * a scalar, a string or a single-dimensional array of these.
 * The table below shows the supported data types and the correspondence
 * to Java and C++ types when DipData objects are passed between C++ and
 * Java applications.
 *<p>
 *
 * <TABLE BORDER="1" summary="DIP Data Types">
 * <TR>
 *    <TD> <b>type name</b>  </TD>
 *    <TD> <b>C++ type</b>   </TD>
 *    <TD> <b>Java type</b>  </TD>
 * </TR>
 * <TR>
 *    <TD> Boolean      </TD>
 *    <TD> bool         </TD>
 *    <TD> boolean      </TD>
 * </TR>
 * <TR>
 *    <TD> Byte         </TD>
 *    <TD> signed char  </TD>
 *    <TD> byte         </TD>
 * </TR>
 * <TR>
 *    <TD> Short </TD>
 *    <TD> short </TD>
 *    <TD> short </TD>
 * </TR>
 * <TR>
 *    <TD> Int  </TD>
 *    <TD> long </TD>
 *    <TD> int  </TD>
 * </TR>
 * <TR>
 *    <TD> Long       </TD>
 *    <TD> long long  </TD>
 *    <TD> long       </TD>
 * </TR>
 * <TR>
 *    <TD> Float </TD>
 *    <TD> float </TD>
 *    <TD> float </TD>
 * </TR>
 * <TR>
 *    <TD> Double </TD>
 *    <TD> double </TD>
 *    <TD> double </TD>
 * </TR>
 * <TR>
 *    <TD> String                 </TD>
 *    <TD> null-terminated string </TD>
 *    <TD> String                 </TD>
 * </TR>
 * </TABLE>
 * For a given type X, the insertion and extraction methods are:
 * <UL>
 * <LI> <b>void insert(X x)</b><br>
 *   This method allows the insertion of a value of type X as a default entry.</LI>
 * <LI> <b>void insert(String tag, X x)</b><br>
 *   This method allows the insertion of a value of type X as the entry named by <code>tag</code>.</LI>
 * <LI><b>X extractX()</b><br>
 *   This method allows the extraction of a value of type X from the default entry.</LI>
 * <LI><b>X extractX(String tag)</b><br>
 *   This method allows the extraction of a value of type X from the entry named by <code>tag</code>.</LI>
 * </UL>
 */
public interface DipData extends Serializable {

  /**
   * Indicates an empty entry. Can be useful for implementations
    * of this interface.
    */
     public static final int TYPE_NULL = 0;

  /**
    * Indicates a <code>boolean</code> value.
    */
     public static final int TYPE_BOOLEAN = 1;

  /**
    * Indicates a <code>boolean</code> array value.
    */
     public static final int TYPE_BOOLEAN_ARRAY = 10;


  /**
    * Indicates a <code>byte</code> value.
    */
     public static final int TYPE_BYTE = 2;
  /**
    * Indicates a <code>byte</code> array value.
    */
     public static final int TYPE_BYTE_ARRAY = 20;


  /**
    * Indicates a <code>short</code> value.
    */
     public static final int TYPE_SHORT = 3;
  /**
    * Indicates a <code>short</code> array value.
    */
     public static final int TYPE_SHORT_ARRAY = 30;


  /**
    * Indicates an <code>int</code> value.
    */
     public static final int TYPE_INT = 4;
  /**
    * Indicates an <code>int</code> array value.
    */
     public static final int TYPE_INT_ARRAY = 40;


  /**
    * Indicates a <code>long</code> value.
    */
     public static final int TYPE_LONG = 5;
  /**
    * Indicates a <code>long</code> array value.
    */
     public static final int TYPE_LONG_ARRAY = 50;


  /**
    * Indicates a <code>float</code> value.
    */
     public static final int TYPE_FLOAT = 6;
  /**
    * Indicates a <code>float</code> array value.
    */
     public static final int TYPE_FLOAT_ARRAY = 60;

  /**
    * Indicates a <code>double</code> value.
    */
     public static final int TYPE_DOUBLE = 7;

  /**
    * Indicates a <code>double</code> array value.
    */
     public static final int TYPE_DOUBLE_ARRAY = 70;

  /**
    * Indicates a <code>String</code> value.
    */
     public static final int TYPE_STRING = 8;
  /**
    * Indicates a <code>String</code> array value.
    */
     public static final int TYPE_STRING_ARRAY = 80;


  /**
   * Returns <code>true</code> if this DipData contains no entries;
    * <code>false</code> otherwise.
    */
     public boolean isEmpty();

  /**
   * Returns the number of entries in this object.
    */
     public int size();

  /**
   * Returns an unordered String array that contains tags of all
   * entries. Returns a zero length array if this object is empty.
    */
     public String[] getTags();

     /**
   * Returns <code>true</code> if this object contains an entry with the
       * specified tag.
       */
    public boolean contains(String tag);

   /**
    * Inserts a float as default value into the DipData object.
    * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
    */
   public void insert(float value) throws TypeMismatch;

   /**
    * Inserts a named float entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
    */
   public void insert(String tag, float value) throws TypeMismatch;

  /**
   * Inserts a double as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(double value) throws TypeMismatch;

  /**
   * Inserts a named double entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, double value) throws TypeMismatch;

  /**
   * Inserts a float array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(float [] value) throws TypeMismatch;

  /**
   * Inserts a named float array entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, float[] value) throws TypeMismatch;

  /**
   * Inserts a double array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(double [] value) throws TypeMismatch;

  /**
   * Inserts a named double array entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, double[] value) throws TypeMismatch;

  /**
   * Inserts a short as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(short value) throws TypeMismatch;

  /**
   * Inserts a named short entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, short value) throws TypeMismatch;

  /**
   * Inserts an short array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(short [] value) throws TypeMismatch;

  /**
   * Inserts a named short array entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, short[] value) throws TypeMismatch;

  /**
   * Inserts an int as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(int value) throws TypeMismatch;

  /**
   * Inserts a named int entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, int value) throws TypeMismatch;

  /**
   * Inserts an int array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(int [] value) throws TypeMismatch;

  /**
   * Inserts a named int array entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, int[] value) throws TypeMismatch;

  /**
   * Inserts a long as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(long value) throws TypeMismatch;

  /**
   * Inserts a named long entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, long value) throws TypeMismatch;

  /**
   * Inserts a long array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(long [] value) throws TypeMismatch;

  /**
   * Inserts a named long array entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, long[] value) throws TypeMismatch;

  /**
   * Inserts a byte as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(byte value) throws TypeMismatch;

  /**
   * Inserts a named byte entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, byte value) throws TypeMismatch;

  /**
   * Inserts a byte array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(byte [] value) throws TypeMismatch;

  /**
   * Inserts a named byte array entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, byte[] value) throws TypeMismatch;

  /**
   * Inserts a boolean as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(boolean value) throws TypeMismatch;

  /**
   * Inserts a named boolean entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, boolean value) throws TypeMismatch;

  /**
   * Inserts a boolean array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(boolean [] value) throws TypeMismatch;

  /**
   * Inserts a named boolean array entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, boolean[] value) throws TypeMismatch;

  /**
   * Inserts a String as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(String value) throws TypeMismatch;

  /**
   * Inserts a named String entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, String value) throws TypeMismatch;

  /**
   * Inserts a String array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is complex or contains a default value which is not of the same type as value.
   */
  public void insert(String [] value) throws TypeMismatch;

  /**
   * Inserts a named String array entry into the DipData object.
   * If the entry already exists, it will be overwritten.
   * @param tag - name of the 'field' the value belongs to.
   * @param value - the value to be written
   * @throws TypeMismatch if entry is simple or contains a tag already exists and its corresponding data is not of the same type as value.
   */
  public void insert(String tag, String[] value) throws TypeMismatch;

  /**
   * Extracts the default value from the DipData object as a float.
   */
  public float extractFloat() throws TypeMismatch;

  /**
   * Returns the <code>float</code> stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>float</code>,
   *         or nothing.
   */
  public float extractFloat(String tag) throws BadParameter, TypeMismatch;

 
  /**
   * Returns the <code>float</code> array stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>float</code> array,
   *         or nothing.
   */
	public float[] extractFloatArray(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>double</code> stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public float[] extractFloatArray() throws TypeMismatch;


	/**
	 * Returns the <code>float</code> array stored in the default data entry.
	 * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
	 */
	public double extractDouble() throws TypeMismatch;

	
	/**
	 * Returns the <code>float</code> array stored in the default data entry.
	  * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>double</code> array,
   *         or nothing.
	 */
	public double extractDouble(String tag) throws BadParameter, TypeMismatch;


  /**
   * Returns the <code>double</code> array stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public double[] extractDoubleArray() throws TypeMismatch;

  /**
   * Returns the <code>double</code> array stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>double</code> array,
   *         or nothing.
   */
  public double[] extractDoubleArray(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>short</code> stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public short extractShort() throws TypeMismatch;

  /**
   * Returns the <code>short</code> stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>short</code>,
   *         or nothing.
   */
  public short extractShort(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>short</code> array stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public short[] extractShortArray() throws TypeMismatch;

  /**
   * Returns the <code>short</code> array stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>short</code> array,
   *         or nothing.
   */
  public short[] extractShortArray(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>int</code> stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public int extractInt() throws TypeMismatch;

  /**
   * Returns the <code>int</code> stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>int</code>,
   *         or nothing.
   */
  public int extractInt(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>int</code> array stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public int[] extractIntArray() throws TypeMismatch;

  /**
   * Returns the <code>int</code> array stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>int</code> array,
   *         or nothing.
   */
  public int[] extractIntArray(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>long</code> stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public long extractLong() throws TypeMismatch;

  /**
   * Returns the <code>long</code> stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>long</code>,
   *         or nothing.
   */
  public long extractLong(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>long</code> array stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public long[] extractLongArray() throws TypeMismatch;

  /**
   * Returns the <code>long</code> array stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>long</code> array,
   *         or nothing.
   */
  public long[] extractLongArray(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>byte</code> stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public byte extractByte() throws TypeMismatch;

  /**
   * Returns the <code>byte</code> stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>byte</code>,
   *         or nothing.
   */
  public byte extractByte(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>byte</code> array stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public byte[] extractByteArray() throws TypeMismatch;

  /**
   * Returns the <code>byte</code> array stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>byte</code> array,
   *         or nothing.
   */
  public byte[] extractByteArray(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>boolean</code> stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public boolean extractBoolean() throws TypeMismatch;

  /**
   * Returns the <code>boolean</code> stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>boolean</code>,
   *         or nothing.
   */
  public boolean extractBoolean(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>boolean</code> array stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public boolean[] extractBooleanArray() throws TypeMismatch;

  /**
   * Returns the <code>boolean</code> array stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>boolean</code> array,
   *         or nothing.
   */
  public boolean[] extractBooleanArray(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>String</code> stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public String extractString() throws TypeMismatch;

  /**
   * Returns the <code>String</code> stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>String</code>,
   *         or nothing.
   */
  public String extractString(String tag) throws BadParameter, TypeMismatch;

  /**
   * Returns the <code>String</code> array stored in the default data entry.
   * @throws TypeMismatch if 1) the object is holding complex data or 2) the value held is not of the required type.
   */
  public String[] extractStringArray() throws TypeMismatch;

  /**
   * Returns the <code>String</code> array stored in the specified data entry.
   * @throws BadParameter
   *         if this DipData does not contain the specified entry
   * @throws TypeMismatch
   *         if the entry contains something other than a <code>String</code> array,
   *         or nothing.
   */
  public String[] extractStringArray(String tag) throws BadParameter, TypeMismatch;


  /**
   * Returns the <code>DipTimestamp</code> of this DipData object. This should represent the time at which the
   * value(s) contained in this object were obtained.
   */
  public DipTimestamp extractDipTime();

  /**
   * Returns the data quality value of this DipData object as an int. The data
   * quality can be one of:
   * @see DipQuality#DIP_QUALITY_BAD
   * @see DipQuality#DIP_QUALITY_GOOD
   * @see DipQuality#DIP_QUALITY_UNCERTAIN
   */
  public int extractDataQuality();
  
  public DipQualityEnum extractDataQualityEnum();
  
  /**
   * Extracts the string that describes the reason for the quality.
   * */
  public String extractQualityString(); 

  /**
    * Returns type code for the value contained in the default entry.
   * The return value can be one of the TYPE_X constants defined in this interface.
    * @see #TYPE_NULL
    * @see #TYPE_BOOLEAN
    * @see #TYPE_BOOLEAN_ARRAY
    * @see #TYPE_BYTE
    * @see #TYPE_BYTE_ARRAY
    * @see #TYPE_SHORT
    * @see #TYPE_SHORT_ARRAY
    * @see #TYPE_INT
    * @see #TYPE_INT_ARRAY
    * @see #TYPE_LONG
    * @see #TYPE_LONG_ARRAY
    * @see #TYPE_FLOAT
    * @see #TYPE_FLOAT_ARRAY
    * @see #TYPE_DOUBLE
    * @see #TYPE_DOUBLE_ARRAY
    * @see #TYPE_STRING
    * @see #TYPE_STRING_ARRAY
   * Will return TYPE_NULL if the object is holding complex data (tag/value pairs)
    */
     public
     int getValueType();

  /**
    * Returns type code for the value contained in the specified entry.
   * The return value can be one of the TYPE_X constants defined in this interface.
   * Will return TYPE_NULL if the tag name passed does not exist.
   * @param tag - the name of the field this method is to find the type of.
    * @see getValueType()
    */
     public
     int getValueType(String tag);

     /**
      * Returns the contents of the DipData as a field name / value map.
      * @return A map associating the field name to its value.
      */
     public Map<String, Object> getValueAsMap() throws Exception;

     /**
      * Indicates whether the DipData package only carries a single value (as opposed to one or more fields, a.k.a. key/value pairs)
      * @return <code>true</code> if the DipData package carries a single value with a default field name.
      */
     public boolean isPrimitive();

     /**
      * Returns the primitive DipData value as a Java Object.
      * @return the primitive value or an empty <code>Optional</code> if the data payload is not a carrying a single value.
      */
     public Optional<Object> getPrimitiveValueAsObject();

}
