package cern.dip;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

/**
 * Utility class to extract any DIP value as an Object, by tag name and value type.
 * @author bcopy
 *
 */
public class DipDataUtil {
        private DipDataUtil() { }
    
	public static final Object extractDipObjectValue(int valueType, DipData dipData,
			String fieldName) throws TypeMismatch, BadParameter {
		Object value = null;
		switch (valueType) {

		case DipData.TYPE_BOOLEAN:
			value = dipData.extractBoolean(fieldName);
			break;

		case DipData.TYPE_BOOLEAN_ARRAY:
			value = dipData.extractBooleanArray(fieldName);
			break;

		case DipData.TYPE_BYTE:
			value = dipData.extractByte(fieldName);
			break;

		case DipData.TYPE_BYTE_ARRAY:
			value = dipData.extractByteArray(fieldName);
			break;

		case DipData.TYPE_SHORT:
			value = dipData.extractShort(fieldName);
			break;

		case DipData.TYPE_SHORT_ARRAY:
			value = dipData.extractShortArray(fieldName);
			break;

		case DipData.TYPE_INT:
			value = dipData.extractInt(fieldName);
			break;

		case DipData.TYPE_INT_ARRAY:
			value = dipData.extractIntArray(fieldName);
			break;

		case DipData.TYPE_LONG:
			value = dipData.extractLong(fieldName);
			break;

		case DipData.TYPE_LONG_ARRAY:
			value = dipData.extractLongArray(fieldName);
			break;

		case DipData.TYPE_FLOAT:
			value = dipData.extractFloat(fieldName);
			break;

		case DipData.TYPE_FLOAT_ARRAY:
			value = dipData.extractFloatArray(fieldName);
			break;

		case DipData.TYPE_DOUBLE:
			value = dipData.extractDouble(fieldName);
			break;

		case DipData.TYPE_DOUBLE_ARRAY:
			value = dipData.extractDoubleArray(fieldName);
			break;

		case DipData.TYPE_STRING:
			value = dipData.extractString(fieldName);
			break;

		case DipData.TYPE_STRING_ARRAY:
			value = dipData.extractStringArray(fieldName);
			break;
		}

		return value;
	}
	
	public static final String convertDipPatternToRegex(String dipTopicPattern) {
	    // Convert the given DIP pattern to a regular expression for easier matching
	    // e.g. dip/acc/LHC/* -> dip\/acc\/LHC/\.*
	    return dipTopicPattern.replace("/", "\\/")
	                          .replace("*", ".*");
	}
	
    public static final Map<String, Object> convertDipDataToObject(DipData data, int maxArraySize) throws TypeMismatch, BadParameter {
        Map<String, Object> newValue = new HashMap<>();

        if (data.isPrimitive()) {
            newValue.put("value", data.getPrimitiveValueAsObject());
            return newValue;
        } else {
            for (String dipTagName : data.getTags()) {

                switch (data.getValueType(dipTagName)) {

                case DipData.TYPE_BOOLEAN:
                case DipData.TYPE_BYTE:
                case DipData.TYPE_DOUBLE:
                case DipData.TYPE_FLOAT:
                case DipData.TYPE_INT:
                case DipData.TYPE_LONG:
                case DipData.TYPE_SHORT:
                case DipData.TYPE_STRING:
                    newValue.put(dipTagName,
                            DipDataUtil.extractDipObjectValue(data.getValueType(dipTagName), data, dipTagName));
                    break;
                case DipData.TYPE_NULL:
                    newValue.put(dipTagName, null);
                    break;

                case DipData.TYPE_BOOLEAN_ARRAY:
                    boolean[] boolArray = data.extractBooleanArray(dipTagName);
                    newValue.put(dipTagName,
                            Arrays.copyOfRange(boolArray, 0, Math.min(boolArray.length, maxArraySize)));
                    break;
                case DipData.TYPE_LONG_ARRAY:
                    long[] longArray = data.extractLongArray(dipTagName);
                    newValue.put(dipTagName,
                            Arrays.copyOfRange(longArray, 0, Math.min(longArray.length, maxArraySize)));
                    break;
                case DipData.TYPE_BYTE_ARRAY:
                    byte[] byteArray = data.extractByteArray(dipTagName);
                    newValue.put(dipTagName,
                            Arrays.copyOfRange(byteArray, 0, Math.min(byteArray.length, maxArraySize)));
                    break;
                case DipData.TYPE_DOUBLE_ARRAY:
                    double[] doubleArray = data.extractDoubleArray(dipTagName);
                    newValue.put(dipTagName, Arrays.copyOfRange(doubleArray, 0,
                            Math.min(doubleArray.length, maxArraySize)));
                    break;
                case DipData.TYPE_FLOAT_ARRAY:
                    float[] floatArray = data.extractFloatArray(dipTagName);
                    newValue.put(dipTagName,
                            Arrays.copyOfRange(floatArray, 0, Math.min(floatArray.length, maxArraySize)));
                    break;
                case DipData.TYPE_INT_ARRAY:
                    int[] intArray = data.extractIntArray(dipTagName);
                    newValue.put(dipTagName,
                            Arrays.copyOfRange(intArray, 0, Math.min(intArray.length, maxArraySize)));
                    break;
                case DipData.TYPE_SHORT_ARRAY:
                    short[] shortArray = data.extractShortArray(dipTagName);
                    newValue.put(dipTagName,
                            Arrays.copyOfRange(shortArray, 0, Math.min(shortArray.length, maxArraySize)));
                    break;
                case DipData.TYPE_STRING_ARRAY:
                    String[] stringArray = data.extractStringArray(dipTagName);
                    newValue.put(dipTagName, Arrays.copyOfRange(stringArray, 0,
                            Math.min(stringArray.length, maxArraySize)));
                    break;

                default:
                    throw new UnsupportedOperationException("DIP Data Type not supported !");
                }

            }

            return newValue;
        }

    }
}
