package cern.dip;
/**
 * Dip (Data Interchange Protocol) is the publish/subscribe interface for the
 * CERN inter-domain communication designed according to the requirements
 * established by the LDIWG (LHC Data Interchange Working Group). <BR><BR>
 *
 * Factory Interface for DipPublication, DipSubscription and DipData objects.
 * <P>
 * Concrete implementations of this interface will require implementation specific
 * parameters such as URL of the broker or naming service.
 * They will normally be implemented as singletons and encapsulate the functionality of the
 * Dip implementation.
 * <P>
 * About DIP:
 * <P>
 * DIP (Data Interchange Protocol) is the publish/subscribe interface for the
 * CERN inter-domain communication designed according to the requirements
 * established by the LDIWG (LHC Data Interchange Working Group). <BR><BR>
 * <P>
 * In this working group has been established that any communication between the
 * independent CERN domains shall be exchanged via DIP. The information namespace
 * shall be partitioned in topics, organised in a hierarchical way with the string
 * "dip" and the string describing the domain name coming first. The remaining part
 * of the topic namespace is under the responsibility of the domain.
 * <p>
 * For instance the energy of the LHC beam accelerator domain topic may be: "dip.acc.LHC.intensity"
 * where "acc" is the accelerator domain.
 */

public interface DipFactory {


  /**
   * Specifies the DNS to be used for the DIP communication. 
   * This setting applies for both the publications and the subscriptions.
   * This method shall be used prior the creation of subscriptions or publications 
   * as these become active immediatly. Further calls will be ignored.
   * @param nodeName name of the DNS to be used.
   */
  void setDNSNode(String nodeName);

  /**
   * Set the write and connection timeout to be used for DIP communication. 
   * @param timeout Timeout in seconds to be applied to new connections.
   */
  void setTimeout(int timeout);


  /**
   * Get the write and connection timeout to be used for DIP communication. 
   */
  int getTimeout();
  
  /**
   * Creates a new DipSubscription object for the topic specified by topicName. Data
   * published on this topic, as well as abnormal events such as disconnections,
   * will be delivered to the listener object.
   * @param topicName name of the subscription topic
   * @param listener is user object which implements
   * @see DipSubscriptionListener interface, used to receive data as well as the state of publication.
   * @throws DipException if the subscription cannot be created.
   */
  DipSubscription createDipSubscription(String topicName, DipSubscriptionListener listener)
      throws DipException;

  /**
   * Creates a new DipPublication object for the topic specified by topicName.
   * This object can then be used to publish new data on the topic. The handler
   * specified by this call is used to inform the Publication developer of any
   * ansynchronous (errors which can not be directly associated with any DipPublication
   * method call).
   * @param topicName name of the subscription topic - must be unique
   * @param handler - used for handling aysnch. DIP errors.
   * @see DipPublicationErrorHandler interface,
   * @throws DipException if the publication cannot be created.
   */
  DipPublication createDipPublication(String topicName, DipPublicationErrorHandler handler)
		throws DipException;


  /**
   * Destroys DipSubscription object previously created by createDipSubscription().
   * @throws DipException if the subscription is not found.
   */
  void destroyDipSubscription(DipSubscription subscription)
      throws DipException;


  /**
   * Destroys DipPublication object previously created by createDipPublication().
   * @throws DipException if the subscription is not found.
   */
  void destroyDipPublication(DipPublication publication)
      throws DipException;

  /**
   * Creates a new DipData object. DipData objects have to be created for publishing
   * complex data.
   */
  DipData createDipData();


  
  /**
  * Create a browser object with which the namespace of 
  * DIP may be viewed.
  */
  DipBrowser createDipBrowser();
}
