package cern.dip;
/**
 * Interface for DIP publications. DipPublication objects are created through the DipFactoryInterface.
 * This object represent's data that is available to subscribers using the DIP protocol and may be used
 * by a publisher to send both primitive data types and DipData objects. The publisher will update a publication
 * according to that publishers criteria (this is usually expected to be when the difference between the last value
 * sent and the current value is sufficiently large).<p>
 * 
 * All send methods take as arguments the value to be sent and the timestamp which will accompany the value.
 *  
 */
public interface DipPublication {
  /**
   * Returns Name of the publication.
   */
  String getTopicName();

  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(double value, DipTimestamp timestamp)
    throws DipException;


  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(double value[], DipTimestamp timestamp)
    throws DipException;


  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(short value, DipTimestamp timestamp)
    throws DipException;
    

  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */  
  void send(short value[], DipTimestamp timestamp)
    throws DipException;


  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(int value, DipTimestamp timestamp)
    throws DipException;
    
    
  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(int value[], DipTimestamp timestamp)
    throws DipException;

   /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(float value, DipTimestamp timestamp)
	throws DipException;

   /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(float value[], DipTimestamp timestamp)
	throws DipException;

  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(long value, DipTimestamp timestamp)
    throws DipException;
    
    
  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */    
  void send(long value[], DipTimestamp timestamp)
    throws DipException;


  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(byte value, DipTimestamp timestamp)
    throws DipException;


  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(byte value[], DipTimestamp timestamp)
    throws DipException;


  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */  
  void send(boolean value, DipTimestamp timestamp)
    throws DipException;
 

  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */  
  void send(boolean value[], DipTimestamp timestamp)
    throws DipException;


  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(String value, DipTimestamp timestamp)
    throws DipException;


  /**
   * Sends publication data
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(String value[], DipTimestamp timestamp)
    throws DipException;


  /**
   * Send complex data object.
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   */
  void send(DipData value, DipTimestamp timestamp)
    throws DipException;
    
  /**
   * Send complex data object with quality information
   * @param value - data to be sent.
   * @param timestamp - time data was obtained.
   * @param dipQuality - the quality of the data
   * @param qualityReason - the reason for the data quality
   */
  public void send(DipData value, DipTimestamp timestamp, int dipQuality,
			String qualityReason) throws DipException;
  /**
   * The quality of the last sent data is set to bad. If this causes a change to the previous 
   * quality then the data (value, quality and timestamp) will be sent to all clients. 
   * This quality is to be used when the data obtained from the source for the publication 
   * is not valid. So instead of introducing bad data into DIP (and thus overwriting the last good
   * value), the quality is set to bad, so still making the last good value available to clients who
   * can make use of it.
   */
  void setQualityBad()
  	throws DipException;
  
  /**
   * As setQualityBad() but enables the user to pass a string to the clients explaining why there is a problem
   * */
  void setQualityBad(String reason)
		throws DipException;


 /**
  * The quality of the last sent data is set to uncertain. If this causes a change to the previous 
  * quality then the data (value, quality and timestamp) will be sent to all clients.
  * This quality is to be used when the data source for the publication is not reachable - thus we
  * can not be certain if the value last sent is still valid.
  */
  void setQualityUncertain()
	  throws DipException;
	  
	  
	/**
	* As setQualityUncertain() but enables the user to pass a string to the clients explaining why there is a problem
	* */
	void setQualityUncertain(String reason)
		throws DipException;
}