package cern.dip.implementation;

import java.io.Serializable;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import cern.dip.BadParameter;
import cern.dip.DipData;
import cern.dip.DipDataUtil;
import cern.dip.DipException;
import cern.dip.DipQualityEnum;
import cern.dip.DipTimestamp;
import cern.dip.TypeMismatch;

/**
 * Interface for DIP data. DipData objects are created by DipFactoryInterfaces and
 * returned to DipSubscriptionListeners.
 */
public class DipDataImp implements DipData, Serializable
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 4385171416542270062L;
	
	/// field name used when dip primitive is sent
	public static final String FIELD_NAME_WHEN_PRMITIVE = "__DIP_DEFAULT__";
	/// Only used on the client side
	private DipQualityEnum dipQuality = DipQualityEnum.GOOD;
	/// Only used on the client side
	private String reasonForQuality = "";  
	/// Only used on the client side  
	private DipTimestamp dipTimestamp;
	   
	  
	private final HashMap<String,Object> fields = new HashMap<>();
	
	
	
	/**
	 * calc how many bytes it would take to store the data. Strings take size() bytes + 1
	 * */
	protected int getCByteSize(){
		Iterator<Object> fieldIt = fields.values().iterator();
		int byteSize = 0;
		while(fieldIt.hasNext()){
			Field field = (Field) fieldIt.next();
			if (field.fieldType == TYPE_STRING){
				String data = (String) field.data;
				byteSize += data.length() + 1;	
			} else if (field.fieldType == TYPE_STRING_ARRAY){
				String [] dataArray = (String[]) field.data;
				for (int arrayInd = 0; arrayInd < field.cardinality; arrayInd++){
					byteSize += dataArray[arrayInd].length() + 1;
				}
			} else{
				byteSize += field.cardinality * getTypeSize(field.fieldType);
			} 
		}
		
		return byteSize;
	}
	
	/**
	 * Write the field to the byte buffer - updating the format string
	 * Format string is formated as follows:-
	 * <name>|<type>|<cardinality>|
	 * */ 
	private String writeToBuffer(Field field, ByteBuffer buffer, String format){
		format += field.fieldName + '|' + field.fieldType + '|' + field.cardinality + '|';
		int startPos = buffer.position();
		switch (field.fieldType){
			case TYPE_BOOLEAN:{
				byte val = (byte)(((Boolean) field.data).booleanValue() == true? 1 : 0);
				buffer.put(val); break;
			}
			case TYPE_BYTE:{
				 buffer.put(((Byte)field.data).byteValue()); break;
			}
			case TYPE_SHORT:{
				buffer.putShort(((Short) field.data).shortValue()); break;
			}
			case TYPE_INT:{
				buffer.putInt(((Integer) field.data).intValue()); break;
			}
			case TYPE_LONG:{
				buffer.putLong(((Long) field.data).longValue()); break;
			}
			case TYPE_FLOAT:{
				buffer.putFloat(((Float) field.data).floatValue()); break;
			}
			case TYPE_DOUBLE:{
				buffer.putDouble(((Double) field.data).doubleValue()); break;
			}
			case TYPE_STRING:{
				String val = (String) field.data;
				String termintatedString = val;
				try{
				buffer.put(val.getBytes("UTF-8"));
				} catch(Exception e){
					return format;
				}
				buffer.put((byte)0);// terminate string
				break;
			}
			case TYPE_BOOLEAN_ARRAY:{
				boolean [] vals = (boolean []) field.data;
				for (int arrayInd = 0; arrayInd < field.cardinality; arrayInd++){
					byte val = (byte)(vals[arrayInd] == true? 1 : 0);
					buffer.put(val);
				}
				break;
			}
			case TYPE_BYTE_ARRAY:{
				byte [] vals = (byte []) field.data;
				buffer.put(vals);
				break;
			}
			case TYPE_SHORT_ARRAY:{
				short [] vals = (short []) field.data;
				for (int arrayInd = 0; arrayInd < field.cardinality; arrayInd++){
					buffer.putShort(vals[arrayInd]);
				}
				break;
			}
			case TYPE_INT_ARRAY:{
				int [] vals = (int []) field.data;
				for (int arrayInd = 0; arrayInd < field.cardinality; arrayInd++){
					buffer.putInt(vals[arrayInd]);
				}
				break;
			}
			case TYPE_LONG_ARRAY:{
				long [] vals = (long []) field.data;
				for (int arrayInd = 0; arrayInd < field.cardinality; arrayInd++){
					buffer.putLong(vals[arrayInd]);
				}
				break;
			}
			case TYPE_FLOAT_ARRAY:{
				float [] vals = (float []) field.data;
				for (int arrayInd = 0; arrayInd < field.cardinality; arrayInd++){
					buffer.putFloat(vals[arrayInd]);
				}
				break;
			}
			case TYPE_DOUBLE_ARRAY:{
				double [] vals = (double []) field.data;
				for (int arrayInd = 0; arrayInd < field.cardinality; arrayInd++){
					buffer.putDouble(vals[arrayInd]);
				}
				break;
			}
			case TYPE_STRING_ARRAY:{
				String [] vals = (String []) field.data;
				for (int arrayInd = 0; arrayInd < field.cardinality; arrayInd++){
					String val = vals[arrayInd];
					try{
						buffer.put(val.getBytes("UTF-8"));
					} catch (Exception e){
						return format;
					}
					buffer.put((byte)0);
				}
				break;
			}
		}
		
		return format;
	}
	
	
	/**
	 * Writes the usr data contained in the object to the start of 
	 * the supplied Data buffer. The format string has the following 
	 * structure : name|type|cardinality|
	 * @param buffer - it is up to the caller to make sure this has sufficient capacity
	 * @return format - the format string describing how the data is packed
	 * */
	protected String writeDataToBuffer(ByteBuffer buffer){
		buffer.position(0);
		Iterator fieldIt = fields.values().iterator();
		String format="";
		while(fieldIt.hasNext()){
			Field field = (Field) fieldIt.next();
			format = writeToBuffer(field, buffer, format);
		}
		return format;
	}
	
	
	
	/**
	 * Extracts a UTF-8 encoded string from the ByteBuffer passed and returns it as 
	 * a unicode'd string.
	 * This method boes NOT preserve the dataBuffer pos varable (it will point to the byte after the 
	 * null terminator of the string extracted).
	 * */
	String extractStringFromBuffer(ByteBuffer buffer){
		int startPos = buffer.position();
		int noUFT8Chars = 0;
		byte uft_8Char;
		do{ // find string size
			uft_8Char = buffer.get();
			noUFT8Chars++; 
		} while (uft_8Char != 0);
		noUFT8Chars--; // do'nt include null char in the buffer
		byte [] uft8_string = new byte[noUFT8Chars];
		buffer.position(startPos);
		for (int charNo = 0; charNo < noUFT8Chars; charNo++){ // copy string
			uft8_string[charNo] = buffer.get();
		}
		uft_8Char = buffer.get(); // skip over null char
		// convert string to uni code
		ByteBuffer uft8_stringBuffer = ByteBuffer.wrap(uft8_string);
		Charset uft_8Converter = Charset.forName("UTF-8");
		CharBuffer unicode_stringBuffer = uft_8Converter.decode(uft8_stringBuffer);
		return unicode_stringBuffer.toString();
	}
	
	
	
	/**
	 * Read data into this object from a ByteBuffer. The format string desribes the structure of the data in the buffer
	 * */
	protected void readInFromBuffer(ByteBuffer dataBuffer, String format) throws TypeMismatch{
		if (format.length() == 0){
			return;
		}
		ByteBuffer buffer = dataBuffer.order(java.nio.ByteOrder.LITTLE_ENDIAN);
		String [] parts = format.split("\\|");
		for (int partIndex = 0; partIndex < parts.length; partIndex+=3){
			String fieldName = parts[partIndex];
			int fieldType = Integer.parseInt(parts[partIndex+1]);
			int fieldCardinality = Integer.parseInt(parts[partIndex+2]);
			
			switch (fieldType){
				case TYPE_BOOLEAN:{
					insert(fieldName, (buffer.get() > 0));break;
				}
				case TYPE_BYTE:{
					insert(fieldName, buffer.get());break;
				}
				case TYPE_SHORT:{
					insert(fieldName, buffer.getShort());break;
				}
				case TYPE_INT:{
					insert(fieldName, buffer.getInt());break;
				}
				case TYPE_LONG:{
					insert(fieldName, buffer.getLong());break;
				}
				case TYPE_FLOAT:{
					insert(fieldName, buffer.getFloat());break;
				}
				case TYPE_DOUBLE:{
					insert(fieldName, buffer.getDouble());break;
				}
				case TYPE_STRING:{
					insert(fieldName, extractStringFromBuffer(buffer));
					break;
				}
				case TYPE_BOOLEAN_ARRAY:{
					boolean [] vals = new boolean[fieldCardinality];
					for (int arrayInd = 0; arrayInd < fieldCardinality; arrayInd++){
						vals[arrayInd] = (buffer.get() > 0);
					}
					insert(fieldName, vals);
					break;
				}
				case TYPE_BYTE_ARRAY:{
					byte [] vals = new byte[fieldCardinality];
					for (int arrayInd = 0; arrayInd < fieldCardinality; arrayInd++){
						vals[arrayInd] = buffer.get();
					}
					insert(fieldName, vals);
					break;
				}
				case TYPE_SHORT_ARRAY:{
					short [] vals = new short[fieldCardinality];
					for (int arrayInd = 0; arrayInd < fieldCardinality; arrayInd++){
						vals[arrayInd] = buffer.getShort();
					}
					insert(fieldName, vals);
					break;
				}
				case TYPE_INT_ARRAY:{
					int [] vals = new int[fieldCardinality];
					for (int arrayInd = 0; arrayInd < fieldCardinality; arrayInd++){
						vals[arrayInd] = buffer.getInt();
					}
					insert(fieldName, vals);
					break;
				}
				case TYPE_LONG_ARRAY:{
					long [] vals = new long[fieldCardinality];
					for (int arrayInd = 0; arrayInd < fieldCardinality; arrayInd++){
						vals[arrayInd] = buffer.getLong();
					}
					insert(fieldName, vals);
					break;
				}
				case TYPE_FLOAT_ARRAY:{
					float [] vals = new float[fieldCardinality];
					for (int arrayInd = 0; arrayInd < fieldCardinality; arrayInd++){
						vals[arrayInd] = buffer.getFloat();
					}
					insert(fieldName, vals);
					break;
				}
				case TYPE_DOUBLE_ARRAY:{
					double [] vals = new double[fieldCardinality];
					for (int arrayInd = 0; arrayInd < fieldCardinality; arrayInd++){
						vals[arrayInd] = buffer.getDouble();
					}
					insert(fieldName, vals);
					break;
				}
				case TYPE_STRING_ARRAY:{
					String [] vals = new String [fieldCardinality];
					for (int arrayInd = 0; arrayInd < fieldCardinality; arrayInd++){
						vals[arrayInd] = extractStringFromBuffer(buffer);
					}
					insert(fieldName, vals);
					break;
				}
			
			}
		}
	}
	
	
	/**
	 * Data object created with a default data quality of good
	 * */
	public DipDataImp()
	{	
	}
	
	public void insert(float value) 
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}


	public void insert(float[] value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

	
  /**
   * Inserts a double as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(double value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts a double array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(double [] value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts a short as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(short value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts an short array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(short [] value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts an int as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(int value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts an int array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(int [] value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts a long as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(long value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts a long array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(long [] value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts a byte as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(byte value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts a byte array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(byte [] value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts a boolean as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(boolean value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts a double array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(boolean [] value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts a String as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(String value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

  /**
   * Inserts a String array as default value into the DipData object.
   * If the default value already exists, it will be overwritten.
   */
	public void insert(String [] value)
	throws TypeMismatch
	{
		insert(FIELD_NAME_WHEN_PRMITIVE, value);
	}

	public String extractString()
		throws TypeMismatch
	{
		try{
		return extractString(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public boolean extractBoolean()
		throws TypeMismatch
	{
		try{
		return extractBoolean(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public byte extractByte()
		throws TypeMismatch
	{
		try{
		return extractByte(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public short extractShort()
		throws TypeMismatch
	{
		try{
		return extractShort(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public int extractInt()
		throws TypeMismatch
	{
		try{
		return extractInt(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public float extractFloat()
		throws TypeMismatch
	{
		try{
		return extractFloat(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public double extractDouble()
		throws TypeMismatch
	{
		try{
		return extractDouble(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public long extractLong()
		throws TypeMismatch
	{
		try{
		return extractLong(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public boolean[] extractBooleanArray()
		throws TypeMismatch
	{
		try{
		return extractBooleanArray(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public byte[] extractByteArray()
		throws TypeMismatch
	{
		try{
		return extractByteArray(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public short[] extractShortArray()
		throws TypeMismatch
	{
		try{
		return extractShortArray(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public int[] extractIntArray()
		throws TypeMismatch
	{
		try{
		return extractIntArray(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public float[] extractFloatArray()
		throws TypeMismatch
	{
		try{
		return extractFloatArray(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public double[] extractDoubleArray()
		throws TypeMismatch
	{
		try{
		return extractDoubleArray(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public long[] extractLongArray()
		throws TypeMismatch
	{
		try{
		return extractLongArray(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	public String[] extractStringArray()
		throws TypeMismatch
	{
		try{
		return extractStringArray(FIELD_NAME_WHEN_PRMITIVE);
		} catch (DipException e){throw new TypeMismatch();}
	}
	
	
	public int extractDataQuality()
	{
		return dipQuality.intValue();
	}
	@Override
	public DipQualityEnum extractDataQualityEnum() {
		return dipQuality;
	}
	
	protected void setDataQuality(int dataQuality){
		this.dipQuality = DipQualityEnum.valueOf(dataQuality);
	}

	public DipTimestamp extractDipTime()
	{
		return dipTimestamp;
	}
	
	protected void setDipTime(DipTimestamp time){
		dipTimestamp = time;
	}
	
	/**
	* Extracts the string that describes the reason for the quality.
	* */
	public String extractQualityString(){
		return reasonForQuality;
	}
	 
	
	protected void setQualityString(String reason){
		reasonForQuality = reason;
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#contains(java.lang.String)
	 */
	public boolean contains(String tag) 
	{
		return fields.containsKey(tag);
	}

	/**
	 * Gets the specified tag after making sure it exists and is of the correct type
	 * Will throw an exception if it does not exist
	 * */
	private Field retrieveField(String tag, int type)
	throws BadParameter, TypeMismatch
	{
		Field field = (Field)lookForTypedField(tag, type);
		if (field == null){
			throw new BadParameter("No such tags " + tag);
		}
		return field;
	}
	
	private Field lookForTypedField(String tag, int type) throws TypeMismatch {
		Field field = (Field)fields.get(tag);
		if (field != null){
			if (field.fieldType!= type){
				throw new TypeMismatch("Tag " + tag + " is of type " + field.fieldType + " not " + type);
			}
		}
		return field;
	}
	
	
	private Field lookForTypedField(String tag, int type, int cardinality) throws TypeMismatch {
		Field field = (Field)fields.get(tag);
		if (field != null){
			if (field.fieldType!= type){
				throw new TypeMismatch("Tag " + tag + " is of type " + field.fieldType + " not " + type);
			}
			if (field.cardinality != cardinality){
				throw new TypeMismatch("Tag " + tag + " of type " + field.fieldType + " has cardinality of " + field.cardinality + " cardinality");
			}
		}
		return field;
	}
	
	
	
	public boolean extractBoolean(String tag)
		throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_BOOLEAN);
		return ((Boolean)field.data).booleanValue();
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractBooleanArray(java.lang.String)
	 */
	public boolean[] extractBooleanArray(String tag)
		throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_BOOLEAN_ARRAY);
		return (boolean[])field.data;
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractByte(java.lang.String)
	 */
	public byte extractByte(String tag)
		throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_BYTE);
		return ((Byte)field.data).byteValue();
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractByteArray(java.lang.String)
	 */
	public byte[] extractByteArray(String tag)
		throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_BYTE_ARRAY);
		return (byte[])field.data;
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractDouble(java.lang.String)
	 */
	public double extractDouble(String tag)
		throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_DOUBLE);
		return ((Double)field.data).doubleValue();
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractDoubleArray(java.lang.String)
	 */
	public double[] extractDoubleArray(String tag)
		throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_DOUBLE_ARRAY);
		return (double[])field.data;
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractFloat(java.lang.String)
	 */
	public float extractFloat(String tag) throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_FLOAT);
		return ((Float)field.data).floatValue();
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractFloatArray(java.lang.String)
	 */
	public float[] extractFloatArray(String tag)
		throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_FLOAT_ARRAY);
		return (float[])field.data;
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractInt(java.lang.String)
	 */
	public int extractInt(String tag)
		throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_INT);
		return ((Integer)field.data).intValue();
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractIntArray(java.lang.String)
	 */
	public int[] extractIntArray(String tag)
		throws BadParameter, TypeMismatch 
	{
		Field field = retrieveField(tag,TYPE_INT_ARRAY);
		return (int[])field.data;
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractLong(java.lang.String)
	 */
	public long extractLong(String tag)
		throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_LONG);
		return ((Long)field.data).longValue();
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractLongArray(java.lang.String)
	 */
	public long[] extractLongArray(String tag)
		throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_LONG_ARRAY);
		return (long[])field.data;
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractShort(java.lang.String)
	 */
	public short extractShort(String tag)
		throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_SHORT);
		return ((Short)field.data).shortValue();
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractShortArray(java.lang.String)
	 */
	public short[] extractShortArray(String tag)
		throws BadParameter, TypeMismatch
	{
		Field field = retrieveField(tag,TYPE_SHORT_ARRAY);
		return (short[])field.data;
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractString(java.lang.String)
	 */
	public String extractString(String tag)
		throws BadParameter, TypeMismatch 
	{
		Field field = retrieveField(tag,TYPE_STRING);
		return (String)field.data;
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#extractStringArray(java.lang.String)
	 */
	public String[] extractStringArray(String tag)
		throws BadParameter, TypeMismatch 
	{
		Field field = retrieveField(tag,TYPE_STRING_ARRAY);
		return (String[])field.data;
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#getTags()
	 */
	public String[] getTags() 
	{
		Set keys = fields.keySet();
		String [] modelArray = new String[0];
		return (String[])keys.toArray(modelArray);
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#getValueType()
	 */
	public int getValueType() {
	
		return getValueType(FIELD_NAME_WHEN_PRMITIVE);
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#getValueType(java.lang.String)
	 */
	public int getValueType(String tag)
	{
		Field field = (Field)fields.get(tag);
		if(field == null)
			return TYPE_NULL;
		return field.fieldType;		
	}


	/**
	 * returns size in bytes of DIP type in bytes
	 * If type is array it will return the size of one of its elements
	 * */
	static public int getTypeSize(int dipType)
	{
		int size = 1;
		if (dipType > 9){
			dipType /= 10; // array type is *10 the enumerated value of its primitive
		}
		switch(dipType)
		{
			case DipData.TYPE_BOOLEAN:
			case DipData.TYPE_BYTE:
			case DipData.TYPE_STRING:
				size = 1;
				break;
			case DipData.TYPE_SHORT:
				size = 2;
				break;
			case DipData.TYPE_INT:
			case DipData.TYPE_FLOAT:
				size = 4;
				break;
			case DipData.TYPE_DOUBLE:
			case DipData.TYPE_LONG:
				size = 8;
				break;
			default:
				size = 1;
				break;	
		}
		return size;
	}


	


	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, boolean)
	 */
	public void insert(String tag, boolean value)
		throws TypeMismatch 
	{
		 if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }

		lookForTypedField(tag, TYPE_BOOLEAN);
		Field field = new Field(new Boolean(value), tag, TYPE_BOOLEAN, 1);
		fields.put(field.fieldName, field);
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, boolean[])
	 */
	public void insert(String tag, boolean[] value) 
	throws TypeMismatch 
	{
        if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }

         if (value == null){
            throw new TypeMismatch("Caller passed null as value");
        }

		lookForTypedField(tag, TYPE_BOOLEAN_ARRAY, value.length);
		boolean []copy = new boolean[value.length];
		for (int i = 0; i < value.length; i++){
			copy[i] = value[i];
		}
		Field field = new Field(copy, tag, TYPE_BOOLEAN_ARRAY, value.length);
		fields.put(field.fieldName, field);
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, byte)
	 */
	public void insert(String tag, byte value) 
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
		lookForTypedField(tag, TYPE_BYTE);
		Field field = new Field(new Byte(value), tag, TYPE_BYTE, 1);
		fields.put(field.fieldName, field);	
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, byte[])
	 */
	public void insert(String tag, byte[] value) 
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
        if (value == null){
            throw new TypeMismatch("Caller passed null as value");
        }
		lookForTypedField(tag, TYPE_BYTE_ARRAY, value.length);
		byte []copy = new byte[value.length];
		for (int i = 0; i < value.length; i++){
			copy[i] = value[i];
		}
		Field field = new Field(copy, tag, TYPE_BYTE_ARRAY, value.length);
		fields.put(field.fieldName, field);	
	}


	public void insert(String tag, double value) 
	throws TypeMismatch 
	{
        if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
		lookForTypedField(tag, TYPE_DOUBLE);
		Field field = new Field(new Double(value), tag, TYPE_DOUBLE, 1);
		fields.put(field.fieldName, field);	
	}


	public void insert(String tag, double[] value) 
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
         if (value == null){
            throw new TypeMismatch("Caller passed null as value");
        }
		lookForTypedField(tag, TYPE_DOUBLE_ARRAY, value.length);
		double []copy = new double[value.length];
		for (int i = 0; i < value.length; i++){
			copy[i] = value[i];
		}
		Field field = new Field(copy, tag, TYPE_DOUBLE_ARRAY, value.length);
		fields.put(field.fieldName, field);	
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, float)
	 */
	public void insert(String tag, float value) 
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
		lookForTypedField(tag, TYPE_FLOAT);
		Field field = new Field(new Float(value), tag, TYPE_FLOAT, 1);
		fields.put(field.fieldName, field);		
	}


	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, float[])
	 */
	public void insert(String tag, float[] value) 
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
        if (value == null){
            throw new TypeMismatch("Caller passed null as value");
        }
		lookForTypedField(tag, TYPE_FLOAT_ARRAY, value.length);
		float []copy = new float[value.length];
		for (int i = 0; i < value.length; i++){
			copy[i] = value[i];
		}
		Field field = new Field(copy, tag, TYPE_FLOAT_ARRAY, value.length);
		fields.put(field.fieldName, field);	
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, int)
	 */
	public void insert(String tag, int value) 
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
		lookForTypedField(tag, TYPE_INT);
		Field field = new Field(new Integer(value), tag, TYPE_INT, 1);
		fields.put(field.fieldName, field);	
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, int[])
	 */
	public void insert(String tag, int[] value) 
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
         if (value == null){
            throw new TypeMismatch("Caller passed null as value");
        }
		lookForTypedField(tag, TYPE_INT_ARRAY, value.length);
		int []copy = new int[value.length];
		for (int i = 0; i < value.length; i++){
			copy[i] = value[i];
		}
		Field field = new Field(copy, tag, TYPE_INT_ARRAY, value.length);
		fields.put(field.fieldName, field);
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, long)
	 */
	public void insert(String tag, long value) 
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
		lookForTypedField(tag, TYPE_LONG);
		Field field = new Field(new Long(value), tag, TYPE_LONG, 1);
		fields.put(field.fieldName, field);
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, long[])
	 */
	public void insert(String tag, long[] value) 
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
        if (value == null){
            throw new TypeMismatch("Caller passed null as value");
        }
		lookForTypedField(tag, TYPE_LONG_ARRAY, value.length);
		long []copy = new long[value.length];
		for (int i = 0; i < value.length; i++){
			copy[i] = value[i];
		}
		Field field = new Field(copy, tag, TYPE_LONG_ARRAY, value.length);
		fields.put(field.fieldName, field);
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, short)
	 */
	public void insert(String tag, short value)
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
		lookForTypedField(tag, TYPE_SHORT);
		Field field = new Field(new Short(value), tag, TYPE_SHORT, 1);
		fields.put(field.fieldName, field);	
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, short[])
	 */
	public void insert(String tag, short[] value) 
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
         if (value == null){
            throw new TypeMismatch("Caller passed null as value");
        }
		lookForTypedField(tag, TYPE_SHORT_ARRAY, value.length);
		short []copy = new short[value.length];
		for (int i = 0; i < value.length; i++){
			copy[i] = value[i];
		}
		Field field = new Field(copy, tag, TYPE_SHORT_ARRAY, value.length);
		fields.put(field.fieldName, field);
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, java.lang.String)
	 */
	public void insert(String tag, String value)
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
          if (value == null){
            throw new TypeMismatch("Caller passed null as value");
        }
		lookForTypedField(tag, TYPE_STRING);
		Field field = new Field(new String(value), tag, TYPE_STRING, 1);
		fields.put(field.fieldName, field);		
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#insert(java.lang.String, java.lang.String[])
	 */
	public void insert(String tag, String[] value) 
	throws TypeMismatch 
	{
         if (tag == null){
            throw new TypeMismatch("Caller passed null as tag name");
        }
        if (value == null){
            throw new TypeMismatch("Caller passed null as value");
        }
		lookForTypedField(tag, TYPE_STRING_ARRAY, value.length);
		String []copy = new String[value.length];
		for (int i = 0; i < value.length; i++){
			copy[i] = new String(value[i]);
		}
		Field field = new Field(copy, tag, TYPE_STRING_ARRAY, value.length);
		fields.put(field.fieldName, field);
	}

	/* (non-Javadoc)
	 * @see cern.dip.DipData#isEmpty()
	 */
	public boolean isEmpty() 
	{
		return fields.isEmpty();
	}
	
	
	
	/* (non-Javadoc)
	 * @see cern.dip.DipData#size()
	 */
	public int size() {
		return fields.size();
	}
	
    @Override
    public Map<String, Object> getValueAsMap() throws Exception {
        Map<String, Object> result = new HashMap<>();

        for (String tag : getTags()) {
            result.put(tag, DipDataUtil.extractDipObjectValue(getValueType(), this, tag));
        }

        return result;
    }

    @Override
    public Optional<Object> getPrimitiveValueAsObject() {
        if (isPrimitive()) {
            return Optional.of(fields.get(FIELD_NAME_WHEN_PRMITIVE));
        } else {
            return Optional.empty();
        }
    }

    @Override
    public boolean isPrimitive() {
        return (fields.size() == 1 && fields.containsKey(FIELD_NAME_WHEN_PRMITIVE));
    }

}	
	
