package cern.dip.implementation;

import java.nio.ByteBuffer;

import cern.dip.DipData;
import cern.dip.DipException;
import cern.dip.DipFactory;
import cern.dip.DipPublicationErrorHandler;
import cern.dip.DipQuality;
import cern.dip.DipTimestamp;

/**
 * Wrapper for the C++ implementation of a DipPublication.
 */
public class DipPublicationImp extends DipPublicationImpBase {

	/**
	 * ID supplied by the C++ implementation
	 * */
	private int publicationID;

	/**
	 * Error handler for the publisher
	 * */
	private DipPublicationErrorHandler errorhandler;

	/**
	 * Used to create DipData objects for sending data of primitive types
	 * */
	DipFactory factory;

	private native int _create(String topicName) throws DipException;

	private native void _destroy(int objectID);

	private native void _setQualityBad(int objectID, String reason)
			throws DipException;

	private native void _setQualityUncertain(int objectID, String reason)
			throws DipException;

	private native void _send(int objectID, ByteBuffer data, String format,
			int noFields, long time_ns, int quality, String qualityReason)
			throws DipException;

	public DipPublicationImp(DipFactory fact, String topicName,
			DipPublicationErrorHandler handler) throws DipException {
		publicationID = -2;
		name = topicName;
		factory = fact;
		errorhandler = handler;
		publicationID = _create(topicName);
	}

	@Override
	protected DipFactory getFactory() {
		return factory;
	}

	/**
	 * Remove C++ implementation if user forgot to call
	 * DipFactory.destroyPublication
	 * */
	public void finalize() {
		if (!isDestroyed() && (publicationID >= 0)) {
			_destroy(publicationID);
		}
	}

	/**
	 * generic send data method used by all send data implementations.
	 * */
	private void sendData(DipData data, DipTimestamp timestamp, int dipQuality,
			String reason) throws DipException {
		validate();
		DipDataImp dataObject = (DipDataImp) data;
		int byteSizeRequired = dataObject.getCByteSize();
		ByteBuffer sharedBuffer = ByteBuffer.allocateDirect(byteSizeRequired);
		sharedBuffer = sharedBuffer.order(java.nio.ByteOrder.LITTLE_ENDIAN);
		String format = dataObject.writeDataToBuffer(sharedBuffer);
		_send(publicationID, sharedBuffer, format, data.size(),
				timestamp.getAsNanos(), dipQuality, reason);
	}

	/**
	 * generic send data method used by all send data implementations.
	 * */
	protected void sendData(DipData data, DipTimestamp timestamp) throws DipException {
		send(data, timestamp, DipQuality.DIP_QUALITY_GOOD, "");
	}

	public void send(DipData value, DipTimestamp timestamp) throws DipException {
		sendData(value, timestamp);
	}

	public void send(DipData value, DipTimestamp timestamp, int dipQuality,
			String qualityReason) throws DipException {
		sendData(value, timestamp, dipQuality, qualityReason);
	}

	public void setQualityBad() throws DipException {
		setQualityBad(null);
	}

	public void setQualityBad(String reason) throws DipException {
		validate();
		_setQualityBad(publicationID, reason);
	}

	public void setQualityUncertain() throws DipException {
		setQualityUncertain(null);
	}

	public void setQualityUncertain(String reason) throws DipException {
		validate();
		_setQualityUncertain(publicationID, reason);
	}

	/**
	 * Delegates to the error handler. This is an entry point for the C++
	 * implementation.
	 * */
	public void handleException(DipException ex) {
		if (this.isDestroyed()) {
			return;
		}
		try {
			errorhandler.handleException(this, ex);
		} catch (Throwable e) {
			System.out.println("PANIC!!: publication " + getTopicName()
					+ "handleException() has thrown exception!");
			e.printStackTrace();
		}
	}

	/**
	 * Mark object as destoyed - will cause the C++ representation to be removed
	 */
	protected void setDestroyed() {
		if (!destroyed) {
			_destroy(publicationID);
			destroyed = true;
		}
	}

}
