package cern.dip.mqtt;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import org.eclipse.paho.client.mqttv3.MqttClient;
import org.eclipse.paho.client.mqttv3.MqttConnectOptions;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttMessage;
import org.eclipse.paho.client.mqttv3.persist.MemoryPersistence;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jmx.export.annotation.ManagedAttribute;
import org.springframework.jmx.export.annotation.ManagedResource;
import org.springframework.stereotype.Component;

import cern.dip.BadParameter;
import cern.dip.Dip;
import cern.dip.DipData;
import cern.dip.DipDataUtil;
import cern.dip.DipFactory;
import cern.dip.DipSubscription;
import cern.dip.DipSubscriptionListener;
import cern.dip.TypeMismatch;
import cern.dip.mqtt.listener.DipUpdateFormatter;
import lombok.extern.slf4j.Slf4j;

@Component
@Slf4j
@ManagedResource(objectName = "bean:name=Subscriber", description = "DIP Subscriber", currencyTimeLimit = 15, persistPolicy = "OnUpdate", persistPeriod = 200, persistLocation = "mqtt", persistName = "subscriber")
public class Subscriber implements DipSubscriptionListener, InitializingBean {

    private DipFactory dipFactory;

    @Autowired
    public SubscriberConfig config;

    MemoryPersistence persistence = new MemoryPersistence();

    MqttConnectOptions connOpts = new MqttConnectOptions();
    
    
    MqttClient mqttClient;
    
    List<DipUpdateFormatter> messageFormatters = new ArrayList<>();

    @Override
    public void afterPropertiesSet() throws Exception {
        dipFactory = Dip.create(config.getName(), config.getDipNs());
        log.info("Created DIP factory to DIP NS '{}'", config.getDipNs());
        mqttClient = new MqttClient(config.getBroker(), config.getName(), persistence);
        log.info("Created MQTT client to broker '{}' with client name '{}'", config.getBroker(), config.getName());
        connOpts.setCleanSession(true);
        connOpts.setAutomaticReconnect(true);
        if(null != config.getUsername() && (!config.getUsername().isEmpty()) ) {
            String display = "(masked)";
            if(config.getPassword().isEmpty()) {
                display = "(empty)";
            }
            log.info("Using MQTT username'{}' password {}", config.getUsername(), display);
            connOpts.setUserName(config.getUsername());
            connOpts.setPassword(config.getPassword().toCharArray());
        }
        start();
    }

    ScheduledExecutorService publishNewDipPublicationsExecutor = Executors.newSingleThreadScheduledExecutor();

    Map<String, DipSubscription> activeDipSubscriptions = new HashMap<>();

    List<DipSubscription> subscriptions = new ArrayList<>();

    List<DipSubscription> connectedSubscriptions = new ArrayList<>();

    Set<String> availableSubscriptions = new HashSet<>();

    public void start() throws MqttException  {
        mqttClient.connect(connOpts);
        final Subscriber parent = this;
        publishNewDipPublicationsExecutor.scheduleAtFixedRate(() -> {
            if (config.publications == null || config.publications.isEmpty()) {
                throw new IllegalStateException(
                        "List of publications to be proxied is empty ! No DIP to MQTT forwarding will take place");
            }
            try {
                // Query all available publications
                for (String pattern : config.publications) {
                    log.debug("Querying DIP pattern {}", pattern);
                    String[] publicationNames = dipFactory.createDipBrowser().getPublications(pattern);
                    log.debug("DIP pattern {} matches {} publication(s)", pattern, publicationNames.length);
                    for (String publicationName : publicationNames) {
                        if (!activeDipSubscriptions.containsKey(publicationName)) {
                            log.info("Subscribing to {} ", publicationName);
                            activeDipSubscriptions.put(publicationName,
                                    dipFactory.createDipSubscription(publicationName, parent));
                        }
                    }

                }

            } catch (Exception t) {
                log.warn("error while querying DIP publications : {}", t);
            }
        }, 1, 60, TimeUnit.SECONDS);
        log.info("Scheduling DIP Namespace watch...");
    }

    @Override
    public void connected(DipSubscription arg0) {
        if (!connectedSubscriptions.contains(arg0)) {
            connectedSubscriptions.add(arg0);
        }
    }

    @Override
    public void disconnected(DipSubscription arg0, String arg1) {
        connectedSubscriptions.remove(arg0);
    }

    @Override
    public void handleException(DipSubscription arg0, Exception arg1) {
        log.error("Received DIP Exception " + arg0.getTopicName(), arg1);
    }

    @Override
    public void handleMessage(final DipSubscription subscription, final DipData dipData) {
        if(log.isTraceEnabled()) {
          log.trace("Forwarding message from {} (ts: {} )", subscription.getTopicName(), dipData.extractDipTime().toString());
        }
        final Subscriber thisSubscriber = this;
        messageFormatters.forEach(formatter -> {
            MqttMessage msg = formatter.prepareMqttMessage(subscription,dipData);
            if(msg != null) {
              thisSubscriber.publishToMQTTBroker(convertDipTopicToMQTTTopic(subscription.getTopicName()), msg);
            }else {
              log.debug("Message formatter {} returned a null MQTT message -> no MQTT publish", formatter);
            }
        });
    }
    
    public void addDipUpdateListener(DipUpdateFormatter listener) {
        messageFormatters.add(listener);
    }
    
    public void publishToMQTTBroker(String topic, MqttMessage msg){
        try {
          mqttClient.publish(topic, msg);
      } catch (Exception e) {
          log.error("Exception while publishing MQTT message ",e);
      }
    }

    public static String convertDipTopicToMQTTTopic(String topicName) {
        return topicName;
    }

    @ManagedAttribute(description = "Number of active DIP subscriptions")
    public int getActiveSubscriptionsCount() {
        return activeDipSubscriptions.size();
    }

    @ManagedAttribute(description = "Set of active DIP subscriptions", currencyTimeLimit = 15)
    public Set<String> getActiveSubscriptionsSet() {
        return new HashSet<>(activeDipSubscriptions.keySet());
    }

    public final Map<String, Object> convertDipDataToObject(DipData data) throws TypeMismatch, BadParameter  {
        Map<String, Object> newValue = new HashMap<>();

        if (data.isPrimitive()) {
            newValue.put("value", data.getPrimitiveValueAsObject());
            return newValue;
        } else {
            for (String dipTagName : data.getTags()) {

                switch (data.getValueType(dipTagName)) {

                case DipData.TYPE_BOOLEAN:
                case DipData.TYPE_BYTE:
                case DipData.TYPE_DOUBLE:
                case DipData.TYPE_FLOAT:
                case DipData.TYPE_INT:
                case DipData.TYPE_LONG:
                case DipData.TYPE_SHORT:
                case DipData.TYPE_STRING:
                    newValue.put(dipTagName,
                            DipDataUtil.extractDipObjectValue(data.getValueType(dipTagName), data, dipTagName));
                    break;
                case DipData.TYPE_NULL:
                    newValue.put(dipTagName, null);
                    break;

                case DipData.TYPE_BOOLEAN_ARRAY:
                    boolean[] boolArray = data.extractBooleanArray(dipTagName);
                    newValue.put(dipTagName,
                            Arrays.copyOfRange(boolArray, 0, Math.min(boolArray.length, config.getMaxArraySize())));
                    break;
                case DipData.TYPE_LONG_ARRAY:
                    long[] longArray = data.extractLongArray(dipTagName);
                    newValue.put(dipTagName,
                            Arrays.copyOfRange(longArray, 0, Math.min(longArray.length, config.getMaxArraySize())));
                    break;
                case DipData.TYPE_BYTE_ARRAY:
                    byte[] byteArray = data.extractByteArray(dipTagName);
                    newValue.put(dipTagName,
                            Arrays.copyOfRange(byteArray, 0, Math.min(byteArray.length, config.getMaxArraySize())));
                    break;
                case DipData.TYPE_DOUBLE_ARRAY:
                    double[] doubleArray = data.extractDoubleArray(dipTagName);
                    newValue.put(dipTagName, Arrays.copyOfRange(doubleArray, 0,
                            Math.min(doubleArray.length, config.getMaxArraySize())));
                    break;
                case DipData.TYPE_FLOAT_ARRAY:
                    float[] floatArray = data.extractFloatArray(dipTagName);
                    newValue.put(dipTagName,
                            Arrays.copyOfRange(floatArray, 0, Math.min(floatArray.length, config.getMaxArraySize())));
                    break;
                case DipData.TYPE_INT_ARRAY:
                    int[] intArray = data.extractIntArray(dipTagName);
                    newValue.put(dipTagName,
                            Arrays.copyOfRange(intArray, 0, Math.min(intArray.length, config.getMaxArraySize())));
                    break;
                case DipData.TYPE_SHORT_ARRAY:
                    short[] shortArray = data.extractShortArray(dipTagName);
                    newValue.put(dipTagName,
                            Arrays.copyOfRange(shortArray, 0, Math.min(shortArray.length, config.getMaxArraySize())));
                    break;
                case DipData.TYPE_STRING_ARRAY:
                    String[] stringArray = data.extractStringArray(dipTagName);
                    newValue.put(dipTagName, Arrays.copyOfRange(stringArray, 0,
                            Math.min(stringArray.length, config.getMaxArraySize())));
                    break;

                default:
                    throw new UnsupportedOperationException("DIP Data Type not supported !");
                }

            }

            return newValue;
        }

    }

}
